#ifndef READ_DIGEST_H
#define READ_DIGEST_H 1

#include "xml_expat.h"

// xoxa_parse_for_digest is implemented in digest.c
// xoxa_digest_{initialize,finalize} are implemented in digest_???.c

/**
 * Parses the content of the input configured into the given parser.
 * This is a convenience wrapper for the sequence:
 *
 *     digest_ctx = xoxa_digest_initialize(parser, digest_name);
 *     parse_status = xoxa_parse(parser);
 *     digest = xoxa_get_digest(parser, digest_ctx, digest_len_ptr);
 *     return digest;
 *
 * or
 *
 *      xoxa_pgp_verify_initialize(ParserState*);
 *      parse_status = xoxa_parse(parser);
 *      return xoxa_get_signature(ParserState*).
 *
 * The list of available digests depends somewhat on the compilation
 * context, but should include for example `"SHA1"`.  If the digest
 * name is "pgp", then do a PGP verification of the XML content.  If
 * it is "_", then scan the content for a <?signature?> PI which
 * indicates the element to digest/verify with "following::*[1]".
 *
 * If there is an error, return NULL; an error message will be
 * available from xoxa_error_message(ParserState*).

 * On successful return, either xoxa_get_signature(ParserState*) or
 * xoxa_get_digest(ParserState* P, unsigned int* lenp) will return non-NULL.
 *
 * @param parser the parser state
 * @param digest_name the name of a digest algorithm
 * @return non-zero on error; zero on success
 */
int xoxa_parse_for_digest(ParserState* parser,
                          const char* digest_name);

/**
 * Initialise the parser for calculating a message digest of the content.
 * The function returns a pointer which must be passed to
 * xoxa_digest_finalize after the parse is complete: this frees
 * memory, and so must be called whether or not the parse was
 * successful.
 *
 * The list of available digests depends somewhat on the compilation
 * context, but should include for example `"SHA1"`.
 *
 * The value <code>void* parser->digest_ctx</code> is available to be
 * set to a digest context object.
 *
 * @param parser the Xoxa parser
 * @param digest_name the name of a digest algorithm
 * @return zero on success, or non-zero on error (possibly with an
 * error message in the parser
 */
int xoxa_digest_initialize(ParserState* parser,
                           const char* digest_name);

/**
 * Tests whether a digest has been initialized.
 *
 * @param parser the parser state
 * @return non-zero (true) if a digest has been initialised
 */
int xoxa_digest_is_initialized(ParserState* parser);

/**
 * Returns some identification of the library providing the digest services.
 * The returned string should not be freed.
 */
const char* xoxa_digest_library_ident(void);

#endif /* READ_DIGEST_H */
