/* This file is part of the Xoxa library: https://bitbucket.org/nxg/xoxa
 * Copyright 2013, 2014, Norman Gray <http://nxg.me.uk>.
 * It is distributed under the terms of the BSD 2-clause Licence;
 * see the file LICENCE for details.
 */
#include <stdio.h>
#include <assert.h>

#define XOXA_PRIVATE_INTERFACE 1
#include "xml_expat.h"
#include "util.h"
#undef XOXA_PRIVATE_INTERFACE
#include "digest.h"

#include "config.h"

// FREE_AND_NULLIFY: free and nullify the argument
#define FREE_AND_NULLIFY(x) if (x != NULL) { \
    free(x);                     \
    x = NULL;                    \
  }


#if HAVE_LIBCRYPTO
// digest_openssl.c needs to be compiled without -Wall on OS X, since from OS X
// 10.9, Apple have marked the entire OpenSSL API as 'deprecated'
// (cf http://rentzsch.tumblr.com/post/33696323211/wherein-i-write-apples-technote-about-openssl-on-os-x;
// on OS X, the module digest_osx.c should be preferred by the configure.ac script).

// Pragmas for clang
#pragma clang diagnostic push
#pragma clang diagnostic ignored "-Wdeprecated-declarations"

// if we HAVE_LIBCRYPTO, we presumably have the headers, too
// (I'm not sure what to do in the alternative case)
#if HAVE_OPENSSL_EVP_H
#include <openssl/evp.h>
#include <openssl/opensslv.h>
#endif

const char* openssl_digest_library_ident(void)
{
    return OPENSSL_VERSION_TEXT;
}

static void openssl_callback(const byte* bytes, size_t len, void* data)
{
    EVP_MD_CTX* mdctx = (EVP_MD_CTX*) data;

    if (! EVP_DigestUpdate(mdctx, bytes, len)) {
        fprintf(stderr, "Can't call digestupdate with bytes %s", bytes);
        exit(1);
    }
}

static void openssl_digest_finalize_(ParserState* parser, void* data)
{

    if (parser->d.digest.ctx == NULL) {
        fatal("digest_openssl: finalize called on non-initialised digest");
    }

    EVP_MD_CTX* mdctx = (EVP_MD_CTX*)parser->d.digest.ctx;

    byte* md_value = NULL;
    int all_ok = 1;
    unsigned int digest_length = 0;

    if ((md_value = (byte*)malloc(EVP_MAX_MD_SIZE)) == NULL) {
        fatal("Can't allocate %d bytes for digest", EVP_MAX_MD_SIZE);
    }
    // this md_value is freed in xoxa_release_parser, after assignment to parser.d.digest.bytes

    if (! EVP_DigestFinal_ex(mdctx, md_value, &digest_length)) {
        xoxa_set_error_message_(parser, 0,
                                "digest_openssl: can't finalise md context");
        all_ok = 0;
    }

    EVP_MD_CTX_cleanup(mdctx);
    FREE_AND_NULLIFY(mdctx);

    if (all_ok) {
        assert(digest_length > 0); // sanity check
        parser->d.digest.bytes = md_value;
        parser->d.digest.len = digest_length;
        assert(parser->d.digest.bytes != NULL);
        parser->flags.digest_type = XOXA_DIGEST_TYPE_MD;
    } else {
        FREE_AND_NULLIFY(md_value);
    }

    assert((all_ok && parser->flags.digest_type == XOXA_DIGEST_TYPE_MD && parser->d.digest.bytes != NULL)
           || (!all_ok && parser->flags.digest_type == XOXA_DIGEST_TYPE_NONE && parser->d.digest.bytes == NULL));

    return;
}

int xoxa_digest_initialize(ParserState* parser,
                           const char* digest_name)
{
    EVP_MD_CTX* mdctx;
    const EVP_MD *md_ident;


    OpenSSL_add_all_digests();

    // The manpage isn't clear, but suggests that this return value
    // points to a static value and does not need to be freed.
    md_ident = EVP_get_digestbyname(digest_name);

    if (md_ident == NULL) {
        xoxa_set_error_message_(parser, 1,
                                "digest_openssl: unknown message digest %s",
                                digest_name);
        // bails out
    }

    MALLOC(mdctx, EVP_MD_CTX);
    // This mdctx is freed in the _finalize function
    
    EVP_MD_CTX_init(mdctx);
    if (! EVP_DigestInit_ex(mdctx, md_ident, NULL)) {
        free(mdctx);
        xoxa_set_error_message_(parser, 1,
                                "digest_openssl: can't initialise md context for digest name %s",
                                digest_name);
        // bails out
    }

    xoxa_push_output_callback(parser, &openssl_callback, mdctx, "openssl-digest");
    xoxa_register_output_callback_finalizer(parser, &openssl_digest_finalize_);

    parser->d.digest.ctx = (void*)mdctx;

    assert(parser->d.digest.ctx != NULL);
    return 0; 
}



#pragma clang diagnostic pop
#endif /* HAVE_LIBCRYPTO */

const char* xoxa_digest_library_ident(void)
{
#if HAVE_LIBCRYPTO
    return openssl_digest_library_ident();
#else
    return NULL;
#endif
}
