/* This file is part of the Xoxa library: https://bitbucket.org/nxg/xoxa
 * Copyright 2013, 2014, Norman Gray <http://nxg.me.uk>.
 * It is distributed under the terms of the BSD 2-clause Licence;
 * see the file LICENCE for details.
 */
#include <CommonCrypto/CommonDigest.h>
#include <string.h> // for memcpy
#include <ctype.h> // for tolower
#include <assert.h>

#define XOXA_PRIVATE_INTERFACE 1
#include "xml_expat.h"
#include "util.h"
#undef XOXA_PRIVATE_INTERFACE

#include "digest.h"

// FREE_AND_NULLIFY: free and nullify the argument
#define FREE_AND_NULLIFY(x) if (x != NULL) { \
    free(x);                     \
    x = NULL;                    \
  }
    

static void sha1_callback(const byte* bytes, size_t len, void* data)
{
    CC_SHA1_Update((CC_SHA1_CTX*)data, (void*)bytes, len);
}

static void sha224_callback(const byte* bytes, size_t len, void* data)
{
    CC_SHA224_Update((CC_SHA256_CTX*)data, (void*)bytes, len);
}

static void sha256_callback(const byte* bytes, size_t len, void* data)
{
    CC_SHA256_Update((CC_SHA256_CTX*)data, (void*)bytes, len);
}

static void sha384_callback(const byte* bytes, size_t len, void* data)
{
    CC_SHA384_Update((CC_SHA512_CTX*)data, (void*)bytes, len);
}

static void sha512_callback(const byte* bytes, size_t len, void* data)
{
    CC_SHA512_Update((CC_SHA512_CTX*)data, (void*)bytes, len);
}

static void md5_callback(const byte* bytes, size_t len, void* data)
{
    CC_MD5_Update((CC_MD5_CTX*)data, (void*)bytes, len);
}

enum digests { invalid, SHA1, SHA2_224, SHA2_256, SHA2_384, SHA2_512, MD5 };
static struct {
    const char* name;
    const enum digests ident;
} all_digests[] = {
    { "md5", MD5 },
    { "sha1", SHA1 },
    { "sha224", SHA2_224 },
    { "sha256", SHA2_256 },
    { "sha384", SHA2_384 },
    { "sha512", SHA2_512 },
};
int n_all_digests = sizeof(all_digests)/sizeof(all_digests[0]);

// The following type matches each of the actual finalizer functions,
// with the second argument cast to void*.
// This lets us avoid having a switch on the digest_ident in both the
// initializer and finalizer function.  This admittedly adds a bit of
// trickiness (including for example this cast), but means they can't
// be out of sync.
typedef int (*cc_generic_finalizer_function)(unsigned char *md, void *c);
typedef struct {
    void* ctx;
    enum digests digest_ident;
    unsigned int digest_length;
    cc_generic_finalizer_function finalizer;
} xoxa_digest_ctx;


//output_callback_end_t
static void osx_digest_finalize_(ParserState* parser, void* data)
{
    if (parser->d.digest.ctx == NULL) {
        fatal("digest_osx: finalize called on non-initialised digest");
    }
    xoxa_digest_ctx* digest_ctx;
    byte* md = NULL;
    unsigned int digest_length = 0;
    int all_ok = 1;

    digest_ctx = (xoxa_digest_ctx*)parser->d.digest.ctx;
    parser->d.digest.ctx = NULL;
    
    digest_length = digest_ctx->digest_length;

    if ((md = malloc(digest_length)) == NULL) {
        fatal("digest_osx: unable to allocate %d bytes for message digest", digest_length);
    }
    // this is freed in xoxa_release_parser, after assignment to parser.d.digest.bytes
    assert(md != NULL);

    (*digest_ctx->finalizer)(md, (void*)digest_ctx->ctx);
    
    FREE_AND_NULLIFY(digest_ctx);

    if (SHOWING_CHATTER(parser)) {
        fprintf(stderr, "osx_digest_finalize_ finalised: digest=");
        {
            int i;
            for (i=0; i<digest_length; i++) {
                fprintf(stderr, " %x", md[i]);
            }
            fprintf(stderr, "\n");
        }
    }

    if (all_ok) {
        // assert(digest_len_ptr != NULL);
        // *digest_len_ptr = digest_length;
        parser->d.digest.bytes = md;
        parser->d.digest.len = digest_length;
        assert(parser->d.digest.bytes != NULL && parser->d.digest.len > 0);
        parser->flags.digest_type = XOXA_DIGEST_TYPE_MD;
    } else {
        // bad-call error earlier
        FREE_AND_NULLIFY(md);
    }

    assert((all_ok && parser->flags.digest_type == XOXA_DIGEST_TYPE_MD && parser->d.digest.bytes != NULL)
           || (!all_ok && parser->flags.digest_type == XOXA_DIGEST_TYPE_NONE && parser->d.digest.bytes == NULL));

    return;
}

int xoxa_digest_initialize(ParserState* parser,
                           const char* digest_name)
{
    xoxa_digest_ctx* digest_ctx;
    enum digests digest_ident = invalid;


    if (parser == NULL) {
        // can't set this to parser->error_message !
        fprintf(stderr, "digest_osx: xoxa_digest_initialize requires initialized parser for digest %s\n", digest_name);
        return 1;
    }

    if (parser->d.digest.ctx != NULL) {
        fprintf(stderr, "digest_osx: xoxa_digest_initialize after initialization -- restarting\n");
        free(parser->d.digest.ctx);
        parser->d.digest.ctx = NULL;
    }

    {
        int i;
        for (i=0; i<n_all_digests; i++) {
            if (strcasecmp(digest_name, all_digests[i].name) == 0) {
                digest_ident = all_digests[i].ident;
                break;
            }
        }
        if (i == n_all_digests) {
            xoxa_set_error_message_(parser, 1,
                                    "digest_osx: unrecognised digest: %s",
                                    digest_name);
            // the above (with second argument 1), does not return, but nonetheless...
            return 1;
        }
    }
    assert(digest_ident != invalid);

#define DO_DIGEST(ctxtype, digestname, callback, length)                \
    {                                                                   \
        if ((digest_ctx->ctx = malloc(sizeof(ctxtype))) == NULL) {      \
            fatal("digest_osx: unable to allocate %d bytes for message context", sizeof(ctxtype)); \
        }                                                               \
        digest_ctx->digest_length = length;                             \
        digest_ctx->finalizer = (cc_generic_finalizer_function)&digestname ## _Final; \
        digestname ## _Init(digest_ctx->ctx);                       \
        xoxa_push_output_callback(parser, &callback, digest_ctx->ctx, \
                                  "osx-digest:" #digestname);        \
    }

    assert(parser->d.digest.ctx == NULL);

    MALLOC(digest_ctx, xoxa_digest_ctx);
    digest_ctx->digest_ident = digest_ident;

    switch (digest_ident) {
      case SHA1:
        DO_DIGEST(CC_SHA1_CTX, CC_SHA1, sha1_callback, CC_SHA1_DIGEST_LENGTH);
        break;
      case SHA2_224:
        DO_DIGEST(CC_SHA256_CTX, CC_SHA224, sha224_callback, CC_SHA224_DIGEST_LENGTH);
        break;
      case SHA2_256:
        DO_DIGEST(CC_SHA256_CTX, CC_SHA256, sha256_callback, CC_SHA256_DIGEST_LENGTH);
        break;
      case SHA2_384:
        DO_DIGEST(CC_SHA512_CTX, CC_SHA384, sha384_callback, CC_SHA384_DIGEST_LENGTH);
        break;
      case SHA2_512:
        DO_DIGEST(CC_SHA512_CTX, CC_SHA512, sha512_callback, CC_SHA512_DIGEST_LENGTH);
        break;
      case MD5:
        DO_DIGEST(CC_MD5_CTX, CC_MD5, md5_callback, CC_MD5_DIGEST_LENGTH);
        break;
      default:
        assert(0);
    }
#undef DO_DIGEST

    xoxa_register_output_callback_finalizer(parser, &osx_digest_finalize_);

    parser->d.digest.ctx = (void*)digest_ctx;

    assert(parser->d.digest.ctx != NULL);

    return 0;
}

#include <sys/sysctl.h>
const char* xoxa_digest_library_ident(void)
{
    static char res[256] = "";

    if (res[0] == '\0') {
        // first time
        char version[256];
        size_t version_size = sizeof(version);
        Buf* digest_list;
        int i;

        (void) sysctlbyname("kern.osrelease", version, &version_size, NULL, 0);

        digest_list = bufinit();

        for (i=0; i<n_all_digests; i++) {
            bufprintf(digest_list, " %s", all_digests[i].name);
        }
        snprintf(res, sizeof(res),
                 "OS X Common Crypto (%s); available digests:%s",
                 version, bufstring(digest_list));

        buffree(digest_list);
    }
    return res;
}
