#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <assert.h>
#include <signal.h>
#include <stdarg.h>
#include <sys/select.h>
#include <ctype.h>

#include "config.h"             // for GPG_PROGRAM
#define XOXA_PRIVATE_INTERFACE 1
#include "util.h" // for buf* functions
#include "pgp.h"
#undef XOXA_PRIVATE_INTERFACE

#ifndef DEBUG
#define DEBUG 0
#endif

#if DEBUG
// force shorter buffers, for debugging, and to stress the code
#define BUFLENGTH 128
#else
#define BUFLENGTH BUFSIZ
#endif

#if DEBUG
#include <fcntl.h>
#include <errno.h>
#endif


extern int errno;


#if DEBUG
// Debugging function, to help work out what FDs are open and which aren't
static char* get_fd_flags(int fd)
{
    int fl;
#define CALL_MAX 3
#define RESP_MAX 5 // at most 4+1 chars in response
    // we allow this to be called three times before it reuses the buffer
    static char call_n = 0;
    static char b[CALL_MAX * RESP_MAX];
    char* p;
    int i;

    p = &b[call_n * RESP_MAX];
    call_n++;
    if (call_n >= CALL_MAX) call_n = 0;

    errno = 0;
    i = 0;
    fl = fcntl(fd, F_GETFL);
    if (fl >= 0) {
        // fine -- these are the flags
        p[i++] = '!';
        if (fl & O_NONBLOCK) p[i++] = 'n';
        if (fl & O_APPEND)   p[i++] = 'a';
        if (fl & O_ASYNC)    p[i++] = 's';
    } else {
        // fl == -1: error, probably EBADF
        if (errno != EBADF) {
            perror("unexpected return from fcntl");
        }
        p[i++] = 'X';
    }
    p[i] = '\0';

    return p;

#undef CALL_MAX
#undef RESP_MAX
}
#endif

// The trailing arguments are of type Channel*, and indicate extra
// input or output channels which are to be associated with the
// process.  The list of arguments must end with NULL.
//
// After this, the ProcessManager owns the Channels and the Bufs
// inside them.  They'll be freed in subprocess_free.
ProcessManager* new_subprocess(char** cmd, ...)
{
    int inpipe[2];              // pipe[0]=read end; pipe[1]=write end
    int i, nchan;
    pid_t child;
    va_list(ap);
    Channel* ch;
    ProcessManager* pm;


    errno = 0;

    if ((pm = malloc(sizeof(ProcessManager))) == NULL) {
        fatal("Can't allocate ProcessManager");
    }

    va_start(ap, cmd);
    nchan = 2;
    for (ch = va_arg(ap, Channel*);
         ch != NULL;
         ch = va_arg(ap, Channel*)) {
        nchan++;
    }
    va_end(ap);
    pm->nchan = nchan;

    if ((pm->ch = malloc(nchan * sizeof(Channel*))) == NULL) {
        fatal("Can't allocate memory for ProcessManager channel");
    }

    pm->ch[0] = new_channel(CHANNEL_READ); // subprocess's stdout -- reading channel
    pm->ch[1] = new_channel(CHANNEL_READ); // stderr
    if (pm->ch[0] == NULL || pm->ch[1] == NULL) {
        goto error_cleanup;
    }

    va_start(ap, cmd);
    for (i=2; i<nchan; i++) {
        pm->ch[i] = va_arg(ap, Channel*);
    }
    va_end(ap);

    if (pipe(inpipe)) {
        perror("inpipe");
        goto error_cleanup;
    }

    child = fork();

    if (child < 0) {
        perror("fork");
        goto error_cleanup;

    } else if (child == 0) {
        // in child
#define CHECK(call, msg) if (call < 0) {        \
            perror(msg);                        \
            goto error_cleanup;                 \
        }

        CHECK(close(inpipe[1]), "child: close parent stdin");
        if (inpipe[0] != 0) {
            CHECK(dup2(inpipe[0], 0), "child: dup stdin");
            CHECK(close(inpipe[0]), "child close child stdin");
        }

        for (i=0; i<nchan; i++) {
            // in the child, close the read/write end that the parent
            // will use to communicate
            CHECK(close(CHANNEL_GET_FD(pm->ch[i])), "child: closing parent end");
        }
        // two special cases:
        if (CHANNEL_GET_CHILD_FD(pm->ch[0]) != 1) {
            int fd = CHANNEL_GET_CHILD_FD(pm->ch[0]);
            CHECK(dup2(fd, 1), "child: dup stdout");
            CHECK(close(fd), "child: close child stdout");
        }
#if !DEBUG
        // if we're in debug mode, then have the child send stderr
        // messages to the same place, rather than to the parent
        if (CHANNEL_GET_CHILD_FD(pm->ch[1]) != 2) {
            int fd = CHANNEL_GET_CHILD_FD(pm->ch[1]);
            CHECK(dup2(fd, 2), "child: dup stderr");
            CHECK(close(fd), "child: close child stderr");
        }
#endif

#if DEBUG
        errno = 0;
            
        fprintf(stderr, "FDs in child: in[%d=%s,%d=%s]",
                inpipe[0], get_fd_flags(inpipe[0]),
                inpipe[1], get_fd_flags(inpipe[1]));
        for (i=0; i<pm->nchan; i++) {
            Channel* ch = pm->ch[i];
            fprintf(stderr, "  %d:[%d=%s,%d=%s]", i,
                    ch->fd[0], get_fd_flags(ch->fd[0]),
                    ch->fd[1], get_fd_flags(ch->fd[1]));
        }
        fprintf(stderr, "; stdin=%s, out=%s, err=%s\n",
                get_fd_flags(0), get_fd_flags(1), get_fd_flags(2));
#endif

        execv(cmd[0], cmd);

        // failed
        perror("execv");
        exit(127);
        
    } else {
        // in parent

        assert(child > 0);

        pm->child = child;

        FD_ZERO(&pm->readable);

        pm->in = inpipe[1];         // write end
        close(inpipe[0]);           // read end

        for (i=0; i<nchan; i++) {
            // in the parent, close the channel that the child will
            // use to communicate
            CHECK(close(CHANNEL_GET_CHILD_FD(pm->ch[i])), "parent: closing child end");
            if (! CHANNEL_FOR_WRITING_P(pm->ch[i])) {
                FD_SET(CHANNEL_GET_FD(pm->ch[i]), &pm->readable);
            }
        }

#if DEBUG
        errno = 0;
        fprintf(stderr, "FDs in parent: in[%d=%s,%d=%s]",
                inpipe[0], get_fd_flags(inpipe[0]),
                inpipe[1], get_fd_flags(inpipe[1]));
        for (i=0; i<pm->nchan; i++) {
            Channel* ch = pm->ch[i];
            fprintf(stderr, "  %d:[%d=%s,%d=%s]", i,
                    ch->fd[0], get_fd_flags(ch->fd[0]),
                    ch->fd[1], get_fd_flags(ch->fd[1]));
        }
        fprintf(stderr, "\n");
#endif

        return pm;
    }
    // can't get here
    assert(0);

#undef CHECK

 error_cleanup:
    fprintf(stderr, "Error creating subprocess\n");
    subprocess_free(pm);
    return NULL;
}

// Cleans up -- kills the child (it's OK if it has already gone),
// and deallocates the ProcessManager structure.
void subprocess_free(ProcessManager* pm)
{
    if (pm == NULL)
        return;

    kill(pm->child, SIGQUIT);   // it's OK if this process no longer exists

    // invalidate, to help detect any reuse after free
    pm->in = -1;

    if (pm->ch != NULL) {
        int i;
        for (i=0; i<pm->nchan; i++) {
            if (pm->ch[i] != NULL) {
                channel_free(pm->ch[i]);
            }
        }
        free(pm->ch);
        pm->ch = NULL;
    }

    pm->in = -1;

    free(pm);
}

// Send the buffer to the subprocess, gathering any pending output.
// If buf is NULL, then close the writing end of the pipe (the child's
// stdin), close the write end of the pipe,
// and read any remaining content from the child, to EOF.
//
// Return zero on success; or non-zero on any problems.
// On the last call (that is, the one with buf=NULL, the return value
// is the subprocess exit code, or -1 if the exit is caused by an
// error uncovered in this function.
int subprocess_send(ProcessManager* pm, const char* buf, size_t buflen)
{
    size_t nwritten = 0;
    char readbuf[BUFLENGTH];
    int work_to_do = 1;
    char last_time_p = 0;
    int highest_fd;
    int status = 0;
    int i;

    struct timeval timeout_val;
    struct timeval* timeout = NULL;


    if (pm == NULL) {
        fatal("subprocess_send: bad call with NULL argument");
    }
    if (pm->in < 0) {
        // this is set by the cleanup code below, to invalidate
        // the structure, before it's deallocated.  If this is true,
        // then the caller is using a structure after freeing.
        // Bad caller.
        fatal("Use of ProcessManager structure after subprocess_send(..., NULL, ...)");
    }

    if (buf == NULL) {
        last_time_p = 1;
        nwritten = buflen = 0;  // nothing to write
        close(pm->in);
    }

    highest_fd = 0;
    if (pm->in > highest_fd) highest_fd = pm->in;
    for (i=0; i<pm->nchan; i++) {
        int fd = CHANNEL_GET_FD(pm->ch[i]);
        if (fd > highest_fd) highest_fd = fd;
    }

    // The logic below is that if there's content in the 'buf' then
    // we'll keep trying until we've written it all out; if this is
    // the last call to this routine (buf is NULL), then we'll keep
    // reading and writing until all of the channels are closed;
    // failing that, we'll keep doing IO while there's content
    // available to write to a writable channel, or content readable.

    while (work_to_do) {
        fd_set readable_now;
        fd_set writable_now;

        int in = pm->in;

        char must_write_p = 0;
        char n_planned = 0;
        char channel_io_happened_p = 0;

        FD_COPY(&pm->readable, &readable_now);
        FD_ZERO(&writable_now);

        if (nwritten < buflen) {
            // still some of this buffer to write
            FD_SET(in, &writable_now);
            must_write_p = 1;
            n_planned++;
        }
        for (i=0; i<pm->nchan; i++) {
            Channel* ch = pm->ch[i];
            int fd = CHANNEL_GET_FD(ch);
            if (CHANNEL_FOR_WRITING_P(ch)) {
                if (ch->buf == NULL) {
                    // if ch->buf is NULL (ie, no Buf has yet been added),
                    // then don't attempt to write it!
                    if (last_time_p) {
                        // ...but this is a mistake
                        bufprintf(pm->ch[1]->buf,
                                  "promised content on channel %d, but nothing there by end of input\n",
                                  i);
                        // this message will be repeated into the
                        // stderr buffer each time round this loop;
                        // fixing that would complicate the logic, and
                        // have little payoff (it's something the user
                        // wants to fix anyway)
                        close(fd);
                    }
                } else {
                    if (ch->nwritten < buflength(ch->buf)) {
                        FD_SET(fd, &writable_now);
                        n_planned++;
                    }
                }
            } else {
                if (FD_ISSET(fd, &readable_now)) {
                    n_planned++;
                }
            }
        }

        if (n_planned == 0) {
            break;              // all done -- JUMP OUT
        }

        if (must_write_p) {
            // we plan to write -- block
            timeout = NULL;
        } else if (last_time_p) {
            // we've finished writing, so now we must wait until the
            // child has finished replying, and closed the fds -- block
            timeout = NULL;
        } else {
            // we've nothing to write, but may have again soon, so if
            // the child doesn't have anything to send back, don't
            // wait -- so don't block, and set a zero timeout
            timeout_val.tv_sec = 0;
            timeout_val.tv_usec = 0;
            timeout = &timeout_val;
        }
#if DEBUG
        fprintf(stderr, "subprocess_send: will write? %s; last time? %s; timeout? %s\n",
                must_write_p ? "yes" : "no",
                last_time_p ? "yes" : "no",
                timeout == NULL ? "block" : "no block");
#endif

        select(highest_fd+1, &readable_now, &writable_now, NULL, timeout);

#if DEBUG
        fprintf(stderr, "subprocess_send: ready(%d", in);
        for (i=0; i<pm->nchan; i++) {
            int fd = CHANNEL_GET_FD(pm->ch[i]);
            char* state;

            if (CHANNEL_FOR_WRITING_P(pm->ch[i])) {
                state = FD_ISSET(fd, &writable_now) ? ">!" : ">.";
            } else {
                state = FD_ISSET(fd, &readable_now) ? "<!" : "<.";
            }
            fprintf(stderr, ",%d%s", fd, state);
        }
        fprintf(stderr, " -> %d) in %s, written %ld/%ld\n",
                highest_fd, 
                (FD_ISSET(in, &writable_now)) ? ">!" : ">.",
                nwritten,
                buflen);
#endif

        if (FD_ISSET(in, &writable_now)) {
            ssize_t nwr = write(in, &buf[nwritten], buflen-nwritten);
#if DEBUG
            fprintf(stderr, "wrote %ld to child\n", nwr);
#endif
            if (nwr >= 0) {
                // normal case
                nwritten += nwr;
            } else {
                // error
                perror("error writing to child");
                close(in);
                status = -1;
            }

        } else {

            for (i=0; i<pm->nchan; i++) {
                Channel* ch = pm->ch[i];
                int fd = CHANNEL_GET_FD(ch);

                if (CHANNEL_FOR_WRITING_P(ch)) {
                    if (FD_ISSET(fd, &writable_now)) {
                        byte* b = bufstring(ch->buf);
                        size_t nw_already = ch->nwritten;
                        ssize_t nwritten = write(fd,
                                                 &b[nw_already],
                                                 buflength(ch->buf)-nw_already);
#if DEBUG
                        fprintf(stderr, "wrote %ld to child through channel %d\n", nwritten, i);
#endif
                        if (nwritten >= 0) {
                            // normal case
                            ch->nwritten += nwritten;
                            if (ch->nwritten == buflength(ch->buf)) {
                                // all done
                                close(fd);
                            }
                        } else {
                            // error
                            perror("error writing to child through channel");
                            close(fd);
                            status = -1;
                        }
                        channel_io_happened_p = 1;
                    }
                } else {
                    // channel for reading
                    Channel* ch = pm->ch[i];
                    int fd = CHANNEL_GET_FD(ch);

                    if (FD_ISSET(fd, &readable_now)) {
                        ssize_t nread = read(fd, readbuf, BUFLENGTH);
#if DEBUG
                        fprintf(stderr, "read %ld from child %d through channel\n", nread, i);
#endif
                        if (nread == 0) {
                            // EOF
                            close(fd);
                            FD_CLR(fd, &pm->readable);
                        } else if (nread < 0) {
                            perror("error reading from child through channel");
                            close(fd);
                            status = -1;
                        } else {
                            // normal case
                            bufstrncpy(ch->buf, (byte*)readbuf, nread);
                        }
                        channel_io_happened_p = 1;
                    }
                } 
            }
        }

        work_to_do = (status == 0) && (must_write_p || channel_io_happened_p);
    }

    if (last_time_p) {
        int exit_status;
#if DEBUG
        fprintf(stderr, "waiting on child %d\n", pm->child);
#endif
        waitpid(-1, &exit_status, 0);
        if (status == 0) {
            // not currently showing an error, so replace it with the
            // exit status of the subprocess
            // fprintf(stderr, "exited? %s; status=%d\n",
            //         WIFEXITED(exit_status) ? "yes" : "no",
            //         WEXITSTATUS(exit_status));
            if (WIFEXITED(exit_status)) {
                status = WEXITSTATUS(exit_status);
            } else {
                status = -1;
            }
        }
    }

    return status;
}

// Called with a Channel which has, or will have, a PGP signature.
// The caller should use subprocess_send to send chunks of the signed
// file to this subprocess, and should then examine the results in the
// ProcessManager object to determine what happened.
ProcessManager* pgp_verify_detached(Channel* sig, Channel* status_info)
{
#define FDBUFLEN 8
    char sig_fd_spec[FDBUFLEN]; // can cope with fd up to 99999
    char status_fd_spec[FDBUFLEN];

    char* cmd[] = { GPG_PROGRAM,
                    "--verify",
                    "--batch",
                    "--enable-special-filenames",
                    "--exit-on-status-write-error",
                    "--status-fd",
                    NULL,       // number from status_info
                    "--",       // end of options
                    NULL,       // will be set to sig_fd_spec
                    "-&0",      // read file from stdin
                    NULL };

    snprintf(status_fd_spec, FDBUFLEN,
             "%d",
             CHANNEL_GET_CHILD_FD(status_info)); // fd of write end
    assert(cmd[6] == NULL);
    cmd[6] = status_fd_spec;

    // we want to set the argument in cmd[6] to be "-&<n>" where <n>
    // is the fd of the read end of the signature channel.  That means
    // that gpg will read that from the Buf which is given to the channel.
    snprintf(sig_fd_spec,  FDBUFLEN,
             "-&%d",
             CHANNEL_GET_CHILD_FD(sig)); // fd of read end
    assert(cmd[8] == NULL);
    cmd[8] = sig_fd_spec;       // was NULL before

#if DEBUG
    {
        char** p;
        fprintf(stderr, "pgp_verify_detached: cmd:");
        for (p = cmd; *p != NULL; p++) {
            fprintf(stderr, " %s", *p);
        }
        fprintf(stderr, "\n");
    }
#endif

    return new_subprocess(cmd, sig, status_info, NULL); // will be NULL on error
#undef FDBUFLEN
}

// Create a new channel, returning NULL on error.
// If the argument is non-zero, this is a writing channel
Channel* new_channel(char writing_p)
{
    Channel* ch;

    if ((ch = malloc(sizeof(Channel))) == NULL) {
        fatal("Can't allocate memory for Channel");
    }

    if (pipe(ch->fd)) {
        perror("new_channel");
        return NULL;
    }

    if (writing_p) {            // non-zero
        ch->buf = NULL;
        ch->nwritten = 0;
    } else {
        ch->buf = bufinit();
        ch->nwritten = -1;
    }

    return ch;
}

void channel_free(Channel* ch)
{
    if (ch != NULL) {
        if (ch->buf != NULL) {
            buffree(ch->buf);
            ch->buf = NULL;
        }
        free(ch);
    }
}

// add a buffer to a writable channel.  Return non-zero on error.
// After this function, the Buf is owned by the Channel
// (in the sense that it will be freed when the Channel is freed).
int channel_add_buffer(Channel* ch, Buf* buf)
{
    int status = 0;

    if (CHANNEL_FOR_WRITING_P(ch)) {
        if (ch->buf == NULL) {
            ch->buf = buf;
            ch->nwritten = 0;
        } else {
            fprintf(stderr, "Channel already has a buffer\n");
            status = 1;
        }
    } else {
        fprintf(stderr, "Can't add a buffer to a readable channel\n");
        status = 1;
    }

    return status;
}
    
// Parse the output sent by gpg to --status-fd.  The structure pointed
// to by 'detail' is modified.  That structure is returned as the
// result, unless there is an error, when NULL is returned.
//
// The 'chatter' buffer is modified by this function (and more or less destroyed).
//
// The results in SignatureDetail share memory with the Buf, and so
// cannot be used after that Buf is freed.
SignatureDetail* gpg_parse_status(Channel* chatter, SignatureDetail* detail)
{
    // Documentation from gnupg installation .../share/doc/gnupg/DETAILS
    //
    // Format of the "--status-fd" output
    // ==================================
    // Every line is prefixed with "[GNUPG:] ", followed by a keyword with
    // the type of the status line and a some arguments depending on the
    // type (maybe none); an application should always be prepared to see
    // more arguments in future versions.
    //
    // ...
    // GOODSIG  <long_keyid_or_fpr>  <username>
    //     The signature with the keyid is good.  For each signature only
    //     one of the codes GOODSIG, BADSIG, EXPSIG, EXPKEYSIG, REVKEYSIG
    //     or ERRSIG will be emitted.  In the past they were used as a
    //     marker for a new signature; new code should use the NEWSIG
    //     status instead.  The username is the primary one encoded in
    //     UTF-8 and %XX escaped. The fingerprint may be used instead of
    //     the long keyid if it is available.  This is the case with CMS
    //     and might eventually also be available for OpenPGP.
    //
    // BADSIG  <long_keyid_or_fpr>  <username>
    //     The signature with the keyid has not been verified okay.  The
    //     username is the primary one encoded in UTF-8 and %XX
    //     escaped. The fingerprint may be used instead of the long keyid
    //     if it is available.  This is the case with CMS and might
    //     eventually also be available for OpenPGP.

    char* cursor = (char*)bufstring(chatter->buf);
    const char* leader = "[GNUPG:]";
    const size_t leader_len = 8;
    char* line;
    char* line_ctx;
    char* word;
    char* word_ctx;

    detail->status = unknown;

    // the parsing here is a little bit precarious; to make it more
    // robust would probably involve a regexp library, or something
    // equally heavyweight.

    line = strtok_r(cursor, "\n\r", &line_ctx);
    while (line != NULL) {
#if DEBUG
        fprintf(stderr, "GPG status: %s\n", line);
#endif
        if (strncmp(line, leader, leader_len) == 0) {
            word = strtok_r(line, " ", &word_ctx); // the string [GNUPG:]
            word = strtok_r(NULL, " ", &word_ctx); // the keyword
            if (strncmp(word, "GOODSIG", 7) == 0) {
                char* p;
                detail->status = good_sig;
                detail->detail.goodsig.keyid = p = strtok_r(NULL, " ", &word_ctx);
                // the the next value is the remainder of the line
                while (! (*p == '\0'))
                    p++;
                p++;
                detail->detail.goodsig.username = p;
            } else if (strncmp(word, "BADSIG", 6) == 0) {
                char* p;
                detail->status = bad_sig;
                detail->detail.badsig.keyid = p = strtok_r(NULL, " ", &word_ctx);
                // the the next value is the remainder of the line
                while (! (*p == '\0'))
                    p++;
                p++;
                detail->detail.badsig.username = p; //strtok_r(NULL, " ", &word_ctx);
            }
        } else {
            fprintf(stderr, "not a gpg line: %s\n", line);
        }
        line = strtok_r(NULL, "\n\r", &line_ctx);
    }

    return detail;
}

// Initialise internal configuration of a SignatureDetail struct.
SignatureDetail* gpg_signature_init(SignatureDetail* details)
{
    details->status = unknown;
    details->stderr = NULL;
    return details;
}

/**
 * Free SignatureDetail contents (note, we do not assume here that the
 * SignatureDetails object itself is malloced, so we do not free it).
 *
 * @param details a pointer to a struct which may have been returned
 * as part of a signature verification process
 */
void gpg_signature_free(SignatureDetail* details)
{
    if (details->stderr != NULL){
        free((void*)details->stderr);
    }
}

typedef struct {
    ProcessManager* pm;
    Channel* status;
} PGPVerifyData;

static void pgp_verify_callback(const byte* bytes, size_t len, void* data)
{
    PGPVerifyData* d = (PGPVerifyData*)data;

#if DEBUG
    fprintf(stderr, "pgp_verify_callback: %ld\n>>>%.*s<<<\n", len, (int)len, bytes);
#endif
    subprocess_send(d->pm, (const char*)bytes, len);
}

static void pgp_signature_callback(const byte* content, size_t len, void* data)
{
    Channel* ch = (Channel*)data;
    Buf* sigbuf = bufinit();

#if DEBUG
    fprintf(stderr, "pgp_signature_callback: %.*s\n", (int)len, content);
#endif
    bufstrncpy(sigbuf, content, len);
    channel_add_buffer(ch, sigbuf);
}

static void xoxa_pgp_verify_finalize_(ParserState* PS, void* data)
{
    PGPVerifyData* d = (PGPVerifyData*)data;

    // finish off sending stuff to the subprocess, irrespective of
    // whether this parse was successful or not
    subprocess_send(d->pm, NULL, 0);

    gpg_parse_status(d->status, &(PS->d.pgp));
    PS->d.pgp.stderr = (const char*)bufstring_copy(d->pm->ch[1]->buf);

    subprocess_free(d->pm);
    free(data);
}

/**
 * Initialise a parser to verify a PGP signature.  On the successful
 * completion of a parser, the results can be retrieved by
 * xoxa_get_signature(ParserState*).
 *
 * @param PS a configured parser state
 * @return zero on success
 */
int xoxa_pgp_verify_initialize(ParserState* PS)
{
    Channel* sig_ch;
    Channel* status_ch;
    PGPVerifyData* verify_data;

    // create a channel through which the signature will be sent to
    // the process; this is initially empty -- we will discover this
    // content during the parse
    sig_ch = new_channel(CHANNEL_WRITE);
    // we'll read status information into the status_ch
    status_ch = new_channel(CHANNEL_READ);

    PS->signature_callback_data = (void*)sig_ch;
    PS->signature_callback = &pgp_signature_callback;

    if ((verify_data = malloc(sizeof(PGPVerifyData))) == NULL) {
        fatal("Can't allocate memory in xoxa_pgp_verify_initialize");
    }
    // this structure is freed in xoxa_pgp_verify_finalize_
    verify_data->pm = pgp_verify_detached(sig_ch, status_ch);
    verify_data->status = status_ch;

    PS->flags.digest_type = XOXA_DIGEST_TYPE_PGP;

    xoxa_push_output_callback(PS,
                              &pgp_verify_callback,
                              (void*)verify_data,
                              "pgp-verify");

    xoxa_register_output_callback_finalizer(PS, &xoxa_pgp_verify_finalize_);

    return 0;
}

/**
 * Given a configured parser, ready to start parsing, verify the PGP
 * signature which we expect to find within the XML file.
 *
 * This is a convenience method, and wraps
 * xoxa_pgp_verify_initialize(ParserState*) and xoxa_get_signature(ParserState*).
 *
 * If the parse is successful, returns the `SignatureDetail` pointer
 * (note that this is dependent on the _parse_ being successful; the
 * verification may or may not be).
 *
 * @param PS a configured parser state
 * @return on error, NULL; on success the filled-in argument `detail`
 */
SignatureDetail* xoxa_parse_for_pgp_verify(ParserState* PS)
{
    int parse_status;

    xoxa_pgp_verify_initialize(PS);

    parse_status = xoxa_parse(PS);

    if (parse_status == 0) {
        // successful parse
        return xoxa_get_signature(PS);
    } else {
        return NULL;
    }
}
