// Use the Expat interface to do an identity transform on an input XML file.
// That is, output the same XML that we parse.
//
// This is a pretty scrappy 'identity transform': We do some limited
// output escaping, but it's quite possible that some odder XML
// documents could confuse this.  This is enough that the identity
// transform, applied to each of the files in the Xoxa test suite,
// produces an XML file that which has the same xoxa-normalised form
// as the input file.
//
// There are two variants, depending on whether USE_XOXA was defined
// as non-zero at compile time.  The difference is that the USE_XOXA
// version uses all of the Xoxa_* 'expat' functions, rather than the
// XML_* expat functions.
//
// Usage:
//    ./identity [-hs] [-o output-file] [xml-file]
//    ./identity_xoxa [-dhs] [-o output-file] [xml-file]
//
// Options:
//    -h: show help
//    -d: show digest of input file
//    -s: show statistics
//    -o: specify output file

#include <stdio.h>
#include <unistd.h>
#include <time.h>
#include <assert.h>
#include <expat.h>
#include <string.h>
#include <fcntl.h>
#include <stdlib.h>
#include <sys/param.h> // for PATH_MAX

#ifndef USE_XOXA
#define USE_XOXA 0
#endif

#if USE_XOXA
#include <xoxa.h>
#include <xoxa_expat.h>
#define DIGEST_NAME "SHA1"
#endif

#if USE_XOXA
#define X_Parser Xoxa_Parser
#define X_ParserCreateNS Xoxa_ParserCreateNS
#define X_SetElementHandler Xoxa_SetElementHandler
#define X_SetCharacterDataHandler Xoxa_SetCharacterDataHandler
#define X_SetExternalEntityRefHandler Xoxa_SetExternalEntityRefHandler
#define X_SetNamespaceDeclHandler Xoxa_SetNamespaceDeclHandler
#define X_SetProcessingInstructionHandler Xoxa_SetProcessingInstructionHandler
#define X_ExternalEntityParserCreate Xoxa_ExternalEntityParserCreate
#define X_SetUserData Xoxa_SetUserData
#define X_GetBuffer Xoxa_GetBuffer
#define X_ParseBuffer Xoxa_ParseBuffer
#define X_GetCurrentLineNumber Xoxa_GetCurrentLineNumber
#define X_GetErrorCode Xoxa_GetErrorCode
#define X_ErrorString Xoxa_ErrorString
#define X_ParserFree Xoxa_ParserFree
#else
#define X_Parser XML_Parser
#define X_ParserCreateNS XML_ParserCreateNS
#define X_SetElementHandler XML_SetElementHandler
#define X_SetCharacterDataHandler XML_SetCharacterDataHandler
#define X_SetExternalEntityRefHandler XML_SetExternalEntityRefHandler
#define X_SetNamespaceDeclHandler XML_SetNamespaceDeclHandler
#define X_SetProcessingInstructionHandler XML_SetProcessingInstructionHandler
#define X_ExternalEntityParserCreate XML_ExternalEntityParserCreate
#define X_SetUserData XML_SetUserData
#define X_GetBuffer XML_GetBuffer
#define X_ParseBuffer XML_ParseBuffer
#define X_GetCurrentLineNumber XML_GetCurrentLineNumber
#define X_GetErrorCode XML_GetErrorCode
#define X_ErrorString XML_ErrorString
#define X_ParserFree XML_ParserFree
#endif

#define NS_SEP (XML_Char)0xff

#define MAXLEVEL 8
typedef struct {
    unsigned int nelem[MAXLEVEL+1];
    unsigned int currlevel;
    unsigned int textlen;
    FILE* outfile;
} Stats;

static const char* progname;

static void Usage(int);
static void parse_file(FILE* input, X_Parser parser);

static void input_file_push(const char*);
static void input_file_pop(void);
static const char* input_file_top(void);

static int ns_stack_pending();
static void ns_stack_push(const XML_Char* uri);
static void ns_stack_pop();
static int ns_stack_size();
static const char* ns_stack_get(int idx);
static int ns_stack_pending();
static int ns_stack_lookup_uri(const char* uri, size_t len);


static void output_escaped_string(const XML_Char* s, int len, FILE* o)
{
    const XML_Char *p = s;
    const XML_Char *p_end = p+len;
    const char *amp_str  = "&amp;";
    const char *lt_str   = "&lt;";
    const char *apos_str = "&apos;";

    assert(o != NULL);

    while (p < p_end) {
        if (*p == '&' || *p == '<' || *p == '\'') {
            if (p > s) {
                fwrite(s, 1, p-s, o);
            }
            switch(*p) {
              case '&':
                fwrite(amp_str, 1, 5, o);
                break;
              case '<':
                fwrite(lt_str, 1, 4, o);
                break;
              case '\'':
                fwrite(apos_str, 1, 6, o);
                break;
              default:
                assert(0);
                break;
            }
            p++;
            s = p;
        } else {
            p++;
        }
    }
    if (p > s) {
        fwrite(s, 1, p-s, o);
    }
}

// crude static allocation
static const char* get_prefixed_name(const XML_Char* name)
{
#define NAMEBUF_LEN PATH_MAX
    static char buf[NAMEBUF_LEN];
        
    const XML_Char* p;

    for (p = name; *p != '\0' && *p != NS_SEP; p++)
        /* nothing */;
    if (*p == NS_SEP) {
        int idx = ns_stack_lookup_uri(name, p-name);
        if (idx < 0) {
            fprintf(stderr, "Can't find declaration of namespace <%.*s>\n",
                    (int)(p-name), name);
            exit(1);
        }
        snprintf(buf, NAMEBUF_LEN, "ns%d:%s", idx, ++p);
    } else {
        assert(*p == '\0');
        snprintf(buf, NAMEBUF_LEN, "%s", name);
    }
    return buf;
#undef NAMEBUF_LEN
}

void start_element_handler(void *userData,
                           const XML_Char *name,
                           const XML_Char **atts)
{
    int i;
    Stats *S = (Stats*)userData;

    if (S->currlevel <= MAXLEVEL) {
        S->nelem[S->currlevel]++;
    }
    S->currlevel++;

    fprintf(S->outfile, "<%s", get_prefixed_name(name));

    for (i=0; atts[i] != NULL; i+=2) {
        fprintf(S->outfile, " %s='", get_prefixed_name(atts[i]));
        output_escaped_string(atts[i+1], strlen(atts[i+1]), S->outfile);
        fprintf(S->outfile, "'");
    }
    for (i = ns_stack_pending(); i < ns_stack_size(); i++) {
        fprintf(S->outfile, " xmlns:ns%d='%s'", i, ns_stack_get(i));
    }
    fprintf(S->outfile, ">");
}

void end_element_handler(void *userData,
                         const XML_Char *name)
{
    Stats *S = (Stats*)userData;

    S->currlevel--;
    fprintf(S->outfile, "</%s>", get_prefixed_name(name));
}

void characterdata_handler(void *userData,
                           const XML_Char *s,
                           int len)
{
    Stats *S = (Stats*)userData;

    S->textlen += len;
    output_escaped_string(s, len, S->outfile);
}

void pi_handler(void *userData,
                const XML_Char *target,
                const XML_Char *data)
{
    Stats *S = (Stats*)userData;
    fprintf(S->outfile, "<?%s %s?>", target, data);
}


void namespace_start_handler(void *userData,
                             const XML_Char *prefix,
                             const XML_Char *uri)
{
    ns_stack_push(uri);
}

void namespace_end_handler(void *userData,
                           const XML_Char *prefix)
{
    ns_stack_pop();
}

int expat_external_entity_handler(X_Parser p,
                                  const XML_Char *context,
                                  const XML_Char *base,
                                  const XML_Char *systemId,
                                  const XML_Char *publicId) 
{
    X_Parser sub_parser;
    FILE* sub_entity;

    int old_cwd = -1;
    const char* bases[2];
    int base_idx;
    int error_p = 0;


    if (systemId == NULL) {
        fprintf(stderr, "%s: external entity with no systemId?!", progname);
        error_p = 1;
        goto clean_up;
    }

    bases[1] = base;
    bases[0] = input_file_top();
    base_idx = 1;

    for (base_idx = 0; base_idx >= 0; base_idx--) {
        if (bases[base_idx] != NULL) {
            const char* base = bases[base_idx];
            const char* slash = strrchr(base, '/');
            if (slash == NULL) {
                // the base URI is just a file -- nothing to do
            } else {
                                // we've found the last slash in the string, so this
                // indicates the directory of the base file
                size_t dirlen = slash - base;
                int chdir_status;
                char* new_cwd = malloc(dirlen + 1);

                if (new_cwd == NULL) {
                    fprintf(stderr, "%s: Can't allocate space for base-URI directory, %s (%lu bytes)",
                            progname, base, dirlen);
                    exit(1);
                }
                memcpy(new_cwd, base, dirlen);
                new_cwd[dirlen] = '\0';

                // open the current working directory, and save its (nonzero) FD
                old_cwd = open(".", O_RDONLY);

                chdir_status = chdir(new_cwd);

                if (chdir_status != 0) {
                    // ooops
                    fprintf(stderr, "%s: Can't cwd to directory %s (base=%s)",
                            progname, new_cwd, base);
                    error_p = 1;
                }
                free(new_cwd);
                if (error_p) {
                    goto clean_up; // JUMP OUT
                }
            }
            base_idx = -1;      // exit loop
        }
    }

    // we don't do anything fancy here, like paying attention to base URLs
    if ((sub_entity = fopen(systemId, "r")) == NULL) {
        fprintf(stderr, "%s: can't open external entity <%s>\n", progname, systemId);
        error_p = 1;
        goto clean_up;
    }

    sub_parser = X_ExternalEntityParserCreate(p, context, NULL);

    input_file_push(systemId);
    parse_file(sub_entity, sub_parser);
    input_file_pop();

 clean_up:
    if (old_cwd >= 0) {
        // change back to the original WD
        fchdir(old_cwd);
        close(old_cwd);
        old_cwd = -1;
    }

    if (error_p) 
        return XML_STATUS_ERROR;
    else
        return XML_STATUS_OK;
}


void parse_file(FILE* input, X_Parser parser)
{
    int keep_reading = 1;
    while (keep_reading) {
        int bytes_read;
        void *buf;

        buf = X_GetBuffer(parser, BUFSIZ);

        if (buf == NULL) {
            fprintf(stderr, "%s: Unable to get Expat buffer\n", progname);
            exit(1);
        }

        if ((bytes_read = fread(buf, 1, BUFSIZ, input)) < 0) {
            fprintf(stderr, "%s: Unable to read XML input\n", progname);
            exit(1);
        }

        if (! X_ParseBuffer(parser, bytes_read, bytes_read == 0)) {
            fprintf(stderr, "%s: Parse error at line %lu: %s\n",
                    progname,
                    X_GetCurrentLineNumber(parser),
                    X_ErrorString(X_GetErrorCode(parser)));
            exit(1);
        }

        keep_reading = (bytes_read > 0);
    }
}



int main(int argc, char** argv)
{
    Stats S;
    int i, ch;
    clock_t starttime, endtime;

    FILE* input = NULL;
    FILE* output = NULL;
    int verbose = 0;
#if USE_XOXA
    int do_digest = 0;
    //void* digest_ctx;
#endif
    X_Parser parser = NULL;


    progname = argv[0];

    while ((ch = getopt(argc, argv, "dhso:")) != -1) {
        switch (ch) {
          case 'd':
#if USE_XOXA
            // the identity program (as opposed to identity_xoxa)
            // recognises the '-d' option, but ignores it (this is so
            // that we can provide this in the benchmarking tests,
            // without error).
            do_digest = 1;
#endif
            break;

          case 's':
            verbose = 1;
            break;

          case 'o':
            if ((output = fopen(optarg, "w")) == NULL) {
                fprintf(stderr, "%s: Can't open file <%s> for output\n",
                        progname, optarg);
                exit(1);
            }
            break;

          case 'h':
            Usage(0);
            break;

          default:
            Usage(1);
        }
    }
    argc -= optind;
    argv += optind;

    switch (argc) {
      case 0:
        input = stdin;
        input_file_push("stdin");
        break;

      case 1:
        if ((input = fopen(argv[0], "r")) == NULL) {
            fprintf(stderr, "%s: Can't open file <%s> for input\n", progname, argv[0]);
            exit(1);
        }
        input_file_push(argv[0]);
        break;

      default:
        Usage(1);
        assert(0);
    }

    parser = X_ParserCreateNS(NULL, NS_SEP);

    if (parser == NULL) {
        fprintf(stderr, "%s: Unable to create Expat parser\n", progname);
        exit(1);
    }

    for (i=0; i<=MAXLEVEL; i++) {
        S.nelem[i] = 0;
    }
    S.currlevel = 0;
    S.textlen = 0;
    S.outfile = (output ? output : stdout);

    assert(input != NULL);
    assert(S.outfile != NULL);

    X_SetElementHandler(parser, start_element_handler, end_element_handler);
    X_SetCharacterDataHandler(parser, characterdata_handler);
    X_SetProcessingInstructionHandler(parser, pi_handler);
    X_SetNamespaceDeclHandler(parser, namespace_start_handler, namespace_end_handler);
    X_SetExternalEntityRefHandler(parser, expat_external_entity_handler);
    X_SetUserData(parser, (void*)&S);

#if USE_XOXA
    if (do_digest) {
        xoxa_digest_initialize(parser, DIGEST_NAME);
    } else {
        xoxa_push_output_callback(parser, &xoxa_null_callback, NULL,
                                  "identity-callback");
    }
#endif

    starttime = clock();

    parse_file(input, parser);
    fputc('\n', S.outfile);

    endtime = clock();

    if (verbose) {
        fprintf(stderr, "Time taken: %gs\n", 1.0*(endtime - starttime)/CLOCKS_PER_SEC);
        fprintf(stderr, "Total text: %u bytes\nNumber of elements at level...\n", S.textlen);
        for (i=0; i<MAXLEVEL; i++) {
            fprintf(stderr, "  %d: %u\n", i, S.nelem[i]);
        }
        fprintf(stderr, "  +: %u\n", S.nelem[MAXLEVEL]);
    }

#if USE_XOXA
    if (do_digest) {
        unsigned char* md;
        size_t digest_length;
        int i;

        //assert(digest_ctx != NULL);
        //md = xoxa_digest_finalize(parser, &digest_length);
        md = xoxa_get_digest(parser, &digest_length);

        printf("%s: %s digest=", progname, DIGEST_NAME);
        for (i=0; i<digest_length; i++) {
            printf(" %02x", md[i]);
        }
        printf("\n");

        fprintf(stderr, "%s: %s digest=", progname, DIGEST_NAME);
        for (i=0; i<digest_length; i++) {
            fprintf(stderr, " %02x", md[i]);
        }
        fprintf(stderr, "\n");
    }
#endif

    X_ParserFree(parser);
}

void Usage(int exit_status)
{
#if USE_XOXA
    const char msg[] = "[-sdh] [-o output-file] [input]\n\
\n\
Using Xoxa.\n\
Options:\n\
    -s: be verbose -- show statistics\n\
    -d: calculate and display the SHA1 digest\n\
    -h: show help\n\
Input defaults to stdin.";
#else
    const char msg[] = "[-sh] [-o output-file] [input]\n\
\n\
Options:\n\
    -s: be verbose -- show statistics\n\
    -h: show help\n\
Input defaults to stdin.";
#endif
    fprintf(stderr, "Usage: %s %s\n", progname, msg);
    exit(exit_status);
}

#define INPUT_FILE_DEPTH 4
static char input_file_stack[INPUT_FILE_DEPTH][PATH_MAX];
static int input_file_ptr = 0;

static void input_file_push(const char* fn)
{
    assert(input_file_ptr >= 0);
    if (input_file_ptr+1 >= INPUT_FILE_DEPTH) {
        fprintf(stderr, "%s: external entity maximum depth (%d) exceeded\n", 
                progname, INPUT_FILE_DEPTH);
        exit(1);
    }
    realpath(fn, input_file_stack[input_file_ptr]);
    //fprintf(stderr, "push %d <%s>\n", input_file_ptr, input_file_stack[input_file_ptr]);
    input_file_ptr++;
    assert(input_file_ptr <= INPUT_FILE_DEPTH);
}

static void input_file_pop(void)
{
    assert(input_file_ptr <= INPUT_FILE_DEPTH);
    if (input_file_ptr == 0) {
        fprintf(stderr, "%s: external entity stack popped off bottom!\n", progname);
        exit(1);
    }
    input_file_ptr--;
    //fprintf(stderr, "popped %d <%s>\n", input_file_ptr, input_file_stack[input_file_ptr]);
    assert(input_file_ptr >= 0);
}

static const char* input_file_top(void)
{
    assert(input_file_ptr > 0 && input_file_ptr <= INPUT_FILE_DEPTH);
    //fprintf(stderr, "top is %d <%s>\n", input_file_ptr, input_file_stack[input_file_ptr-1]);
    return input_file_stack[input_file_ptr-1];
}

// The following is a very crude stack of URI->namespace mappings
#define NS_STACK_SIZE 8
static char* ns_stack[NS_STACK_SIZE];
static int ns_stack_pending_ = 0;
static int ns_stack_size_ = 0;
static void ns_stack_push(const XML_Char* uri)
{
    assert(ns_stack_size_+1 < NS_STACK_SIZE);
    if ((ns_stack[ns_stack_size_] = malloc(strlen(uri)+1)) == NULL) {
        fprintf(stderr, "unable to allocate %ld bytes for namespace stack\n", (strlen(uri)+1));
        exit(1);
    }
    strcpy(ns_stack[ns_stack_size_], uri);
    ++ns_stack_pending_;
    ++ns_stack_size_;
}
static void ns_stack_pop()
{
    assert(ns_stack_size_ > 0);
    --ns_stack_size_;
    assert(ns_stack[ns_stack_size_] != NULL);
    free(ns_stack[ns_stack_size_]);
}

static int ns_stack_size()
{
    return ns_stack_size_;
}

static const char* ns_stack_get(int idx)
{
    assert(idx < ns_stack_size_);
    return ns_stack[idx];
}

// Return the index of the last 'pending' namespace prefix
static int ns_stack_pending()
{
    int t = ns_stack_size_ - ns_stack_pending_;
    assert(t >= 0);
    ns_stack_pending_ = 0;
    return t;
}

// Return the index of the given URI
static int ns_stack_lookup_uri(const char* uri, size_t len)
{
    int i;
    for (i=ns_stack_size_ - 1; i >= 0; i--) {
        if (strncmp(uri, ns_stack[i], len) == 0)
            return i;
    }
    return -1;
}
