#include <assert.h>
#include <string.h>
#include <stdlib.h>
#include <sys/syslimits.h>

// There is some remaining overlap between the tests here, and the tests
// in ../../../tests/testcases/misc/*.xml.  Tests that certain things
// fail can be only here; having tests in .../*.xml means that we
// guarantee the same behaviour from the C and Java libraries
// (some of the tests are/were tests of XML parsing,
// which are relics of when this library contained its own basic XML
// parser, which was fairly early on replaced by Expat).
//
// Most of the remaining tests, therefore, are tests of the interface
// with the parsing functions in the library.


#define XOXA_PRIVATE_INTERFACE 1
// ...so we get the buf*() functions, too
#include "util.h"
#include "pgp.h"
#undef XOXA_PRIVATE_INTERFACE

#include "xoxa.h"

#include "lcut/lcut.h"
// FAIL_UNLESS_LCUT_ASSERT is like LCUT_ASSERT, but it returns immediately on failure
// (also it has variadic message)
#define FAIL_UNLESS_LCUT_ASSERT(tc, condition, msg, ...)        \
    do {                                                        \
        char buf[128];                                          \
        snprintf(buf, 128, msg, __VA_ARGS__);                   \
        LCUT_ASSERT(tc, buf, condition);                        \
        if ((tc)->status == TEST_CASE_FAILURE) { return; }      \
    }                                                           \
    while (0)



// Helper functions for checking results
//
// Register 'gathering_callback' as the callback function, and then
// call 'gather_parse_output'.  The individual records will then be in
// lines[i], where 0 <= i < nlines.  Then call 'reset_collector'.
#define NLINES 64
static char* lines[NLINES];
static int nlines;
static Buf* linebuf = NULL;

void gathering_callback(const byte* bytes, size_t len, void* data)
{
    // the array 'bytes' is a sequence of "foo\r\nbar\r\n...\r\n".
    // However they don't necessarily arrive as complete lines, so
    // simply concatenate them.

    if (linebuf == NULL) {
        // first time
        linebuf = bufinit();
    }

    if (len == 0) {
        // null terminated
        bufstrcpy(linebuf, bytes);
    } else {
        bufstrncpy(linebuf, bytes, len);
    }
}
static void gather_parse_output(void)
{
    // the lines[] array points into this 'output' array, so don't
    // free it at the end of this function.
    static byte* output = NULL;
    byte* p;

    if (output != NULL) {
        free(output);
    }

    output = bufstring_copy(linebuf);
    p = output;

    for (nlines=0; *p != '\0'; nlines++) {
        if (nlines >= NLINES) {
            fprintf(stderr, "too many lines in unit-tests: increase NLINES\n");
            exit(1);
        }

        lines[nlines] = (char*)p;

        // find the next end-of-line
        while (*p != 0 && *p != '\r') {
            p++;
        }
        if (*p == '\0') {
            // malformed
            fprintf(stderr, "Malformed input in test (no CR): %s\n", output);
            exit(1);
        }
        // pointing at '\r'
        if (p[1] != '\n') {
            // malformed
            fprintf(stderr, "Malformed input in test (no LF): %s\n", output);
            exit(1);
        }
        p[0] = '\0';
        p += 2;
        // pointing at next 'line' or \0
    }
}
static void reset_collector(void)
{
    if (linebuf == NULL) {
        linebuf = bufinit();
    } else {
        bufreset(linebuf);
    }
}
        
    
void setup_test_callbacks(ParserState* PS)
{
    xoxa_push_output_callback(PS, gathering_callback, NULL, "gathering-callback");
    reset_collector();
}

#define ERRBUFLEN 256
char errbuf[ERRBUFLEN+1];

// Test a number of failing cases -- check that they're detected
void tc_check_error_parse(lcut_tc_t *tc, void *data)
{
    InputSource* src;
    ParserState* PS;
    const char* tests[] = {
        "", // empty!
        "<e>\v</e>",            // invalid character
        "<e>\b</e>",            // invalid character
        "<e<e1/>/>",            // invalid '<' in tag and ... more
        "<e>there</e",          // missing >
        "<e a1=v1>there</e>",   // no quotes
        "<e a1='v1\">foo</e>",  // mismatched quotes
        "<e p: a1='v1'/>",      // malformed name
        "<e p :a1='v1'/>",      // malformed name
        "<e a1='v&'/>",         // ampersand in attribute
        "<e a1='v& '/>",        // ampersand in attribute
        "<e a1='v<'/>",         // &lt; in attribute
        "<e>there</foo>",       // mismatched tags
        "<e>there",             // unclosed element
        "<e><!--x---></e>",     // bad comment close
        "<e><!--x-></e>",       // unclosed comment
        "<e><!--x----y--></e>", // two comments within markup
        "<e>text<![ CDATA[foo]]></e>", // incorrectly started
        "<e>text<![CDATA [foo]]></e>", // incorrectly started
        "<e>text<![CDATA[foo]></e>",   // incorrectly closed
        "<e><![ELSE[foo]]></e>",       // not CDATA
        "<e>&</e>",             // unescaped ampersand
        "<e><</e>",             // unescaped angle bracket
        "<e>&;</e>",            // empty reference
        "<e>ent &abc",          // EOF in entity reference
        "<e>&foo;</e>",         // undefined entity (should this be an error?)
        "<e>&#3f;</e>",         // hex entity without leading 'x'
        "<e>&#x1g;</e>",        // malformed hex number
        "x<e/>",                // text outside of document element
        "<e/> x",               // text outside of document element
    };
    int ntests = sizeof(tests)/sizeof(tests[0]);
    int i;

    for (i=0; i<ntests; i++) {
        int error_p = 0;

        src = xoxa_new_input_source_from_string(tests[i]);
        PS = xoxa_new_parser(src);
        setup_test_callbacks(PS);

        snprintf(errbuf, ERRBUFLEN, "parse didn't fail: %s", tests[i]);
        error_p = xoxa_parse(PS);
        LCUT_ASSERT(tc, errbuf, error_p);

        xoxa_release_parser(PS);
    }
}

static void assert_strings_match(lcut_tc_t *tc,
                                 const char* xml,
                                 ParserFlags parser_flags,
                                 char** expected,
                                 int ne,
                                 char* msg,
                                 int line,
                                 const char* func)
{
    InputSource* src = xoxa_new_input_source_from_string(xml);
    ParserState* parser = xoxa_new_parser(src);

    if (parser_flags) {
        xoxa_parser_flags(parser, parser_flags);
    }

    setup_test_callbacks(parser);

    if (xoxa_parse(parser) == 0) {
        int i;
        gather_parse_output();

#if 0
        for (i=0; i<nlines; i++) {
            fprintf(stderr, "%d: %s\n", i, lines[i]);
        }
#endif

        //LCUT_INT_EQUAL(tc, ne, nextline);
        lcut_int_equal(tc, ne, nlines, line, func, __FILE__);
        for (i=0; i<ne; i++) {
            if (strcmp(lines[i], expected[i]) != 0) {
                snprintf(errbuf, ERRBUFLEN, "%s, item %d: actual <%s>, expected <%s>\n",
                         msg,
                         i, lines[i], expected[i] );
                //LCUT_ASSERT(tc, "failed to match", 0);
                lcut_assert(tc, errbuf, 0, line, func, __FILE__);
                break;
            }
        }

    } else {
        //LCUT_ASSERT(tc, "%s: failed to parse string", 0);
        snprintf(errbuf, ERRBUFLEN, "%s: failed to parse string", msg);
        lcut_assert(tc, errbuf, 0, line, func, __FILE__);
    }

    xoxa_release_parser(parser);
}

// Check that a simple parse doesn't blow up
void tc_check_simple_parse(lcut_tc_t *tc, void *data)
{
    const char* testxml =
         "<e1 a1='v1'>\n"
         "  <e2 xmlns:p1=\"ns1\">\n"
         "    <e3 a3='v3' p1:a4='v4' xmlns='ns2'>\n"
         "      <p1:e4 a5='v5'>text</p1:e4>\n"
         "      <e5 a6='v6'>text2</e5>\n"
         "      <p1:e6 p1:a7='v7' a8='v8'/>\n"
         "      <e7 p1:a9='v9' a10='v10'/>\n"
         "    </e3>\n"
         "    <e8 a11='v11'/>\n"
         "    <e9 a12='v12' a14='v14' p1:a15='v15' a13='v13'/>\n" // requires reordering
         "  </e2>\n"
         "</e1>\n";

    char* expected[] = { 
        "Aa1 CDATA v1",
        "(e1",
        //"-\\n  ",
        //"Mp1 ns1",
        "(e2",
        //"-\\n    ",
        //"M ns2",
        "Aa3 CDATA v3",
        "Bns1 a4 CDATA v4",
        "[ns2 e3",
        //"-\\n      ",
        "Aa5 CDATA v5",
        "[ns1 e4",
        "-text",
        "]ns1 e4",
        //"-\\n      ",
        "Aa6 CDATA v6",
        "[ns2 e5",
        "-text2",
        "]ns2 e5",
        //"-\\n      ",
        "Aa8 CDATA v8",
        "Bns1 a7 CDATA v7",
        "[ns1 e6",
        "]ns1 e6",
        //"-\\n      ",
        "Aa10 CDATA v10",
        "Bns1 a9 CDATA v9",
        "[ns2 e7",
        "]ns2 e7",
        //"-\\n    ",
        "]ns2 e3",
        //"m",
        //"-\\n    ",
        "Aa11 CDATA v11",
        "(e8",
        ")e8",
        //"-\\n    ",
        "Aa12 CDATA v12",
        "Aa13 CDATA v13",
        "Aa14 CDATA v14",
        "Bns1 a15 CDATA v15",
        "(e9",
        ")e9",
        //"-\\n  ",
        ")e2",
        //"mp1",
        //"-\\n",
        ")e1",
    };
    int ne = sizeof(expected)/sizeof(expected[0]);
    // expected_full is as above, but with the extra items included
    char* expected_full[] = { 
        "Aa1 CDATA v1",
        "(e1",
        "-\\n  ",
        "Mp1 ns1",
        "(e2",
        "-\\n    ",
        "M ns2",
        "Aa3 CDATA v3",
        "Bns1 a4 CDATA v4",
        "[ns2 e3",
        "-\\n      ",
        "Aa5 CDATA v5",
        "[ns1 e4",
        "-text",
        "]ns1 e4",
        "-\\n      ",
        "Aa6 CDATA v6",
        "[ns2 e5",
        "-text2",
        "]ns2 e5",
        "-\\n      ",
        "Aa8 CDATA v8",
        "Bns1 a7 CDATA v7",
        "[ns1 e6",
        "]ns1 e6",
        "-\\n      ",
        "Aa10 CDATA v10",
        "Bns1 a9 CDATA v9",
        "[ns2 e7",
        "]ns2 e7",
        "-\\n    ",
        "]ns2 e3",
        "m",
        "-\\n    ",
        "Aa11 CDATA v11",
        "(e8",
        ")e8",
        "-\\n    ",
        "Aa12 CDATA v12",
        "Aa13 CDATA v13",
        "Aa14 CDATA v14",
        "Bns1 a15 CDATA v15",
        "(e9",
        ")e9",
        "-\\n  ",
        ")e2",
        "mp1",
        "-\\n",
        ")e1",
    };
    int ne_full = sizeof(expected_full)/sizeof(expected_full[0]);

    assert_strings_match(tc, testxml, 0,
                         expected, ne,
                         "Simple parse",
                         __LINE__, __FUNCTION__);
    assert_strings_match(tc, testxml, XOXA_REPORT_ALL,
                         expected_full, ne_full,
                         "Simple parse (reporting all)",
                         __LINE__, __FUNCTION__);
}

void tc_check_entity_expansion(lcut_tc_t *tc, void *data)
{
    const char* testxml =
        "<e>Text &lt;&gt;&amp;&apos;&quot;"
         "<e1 a1='v&amp;&lt;&gt;>'>&#x40;&#37;&#13;</e1>" // this &#13; should be the only way of getting a \r into the string
         "</e>";
    char* expected[] = {
        "(e",
        "-Text <>&'\"",
        "Aa1 CDATA v&<>>",
        "(e1",
        "-@% ",
        ")e1",
        ")e",
    };
    int ne = sizeof(expected)/sizeof(expected[0]);

    assert_strings_match(tc, testxml, 0, expected, ne,
                         "Entity expansion",
                         __LINE__, __FUNCTION__);
}

// With this and the following test, we should add further cases if
// and when we add support for non-default non-charref entities
// which may contain whitespace.
void tc_check_line_end_normalisation(lcut_tc_t *tc, void *data)
{
    const char* testxml =
         "<doc>1\n"
         "2\r\n"
         "3\n\r"
         // XXX the next two will need to wait until I add unicode support
         // to the tokenizer
         //"4\u0085" // NEL, XML 1.1 only
         //"5\u2028" // Unicode line-separator
         "6\r"
         "7&#xd;"               // expands to \r
         "8"
         "</doc>";
    char* expected[] = {
        "(doc",
        "-1 2 3 6 7 8",
        ")doc",
    };
    int ne = sizeof(expected)/sizeof(expected[0]);

    assert_strings_match(tc, testxml, 0, expected, ne,
                         "Line-end normalisation",
                         __LINE__, __FUNCTION__);   
}

void tc_check_whitespace_normalisation(lcut_tc_t *tc, void *data)
{
    const char* testxml =
         "<doc"
        // Attribute Value Normalisation (XML 1.1, Sect 3.3.3) means
        // that the \n, \r, \t are replaced by ' ' in _both_ the
        // xoxa-normalised version and the unnormalised version.
        // The Character References, however, survive into the latter one.
         " a1='  v1\n\t\r x '"
         " a2='  &#xd;x\ry\n\rz&#xa;&#x9; '"
         ">\n"
        "  <e1>  1 \n"
        "  2 \n\r"
        "3\n"
        "\n\r"
        "\r"
        "   4\r"
        "</e1>"
        "<e2>5</e2>" // no whitespace
        "\n\r\n"
        "  </doc>";
    char* expected_normalised[] = {
        "Aa1 CDATA  v1 x ",
        "Aa2 CDATA  x y z ",
        "(doc",
        "(e1",
        "- 1 2 3 4 ",
        ")e1",
        "(e2",
        "-5",
        ")e2",
        ")doc",
    };
    int ne = sizeof(expected_normalised)/sizeof(expected_normalised[0]);
    char* expected_unnormalised[] = {
        "Aa1 CDATA   v1    x ",
        "Aa2 CDATA   \\rx y  z\\n\t ",
        "(doc",
        "-\\n  ",
        "(e1",
        // XML 1.1 Sect. 2.11 means that the \r in the input are
        // translated into \n
        "-  1 \\n  2 \\n\\n3\\n\\n\\n\\n   4\\n",
        ")e1",
        "(e2",
        "-5",
        ")e2",
        "-\\n\\n  ",
        ")doc",
    };
    int ne_not = sizeof(expected_unnormalised)/sizeof(expected_unnormalised[0]);

    assert_strings_match(tc, testxml, 0, expected_normalised, ne,
                         "Whitespace normalisation",
                         __LINE__, __FUNCTION__);   
    assert_strings_match(tc, testxml, XOXA_REPORT_ALL, expected_unnormalised, ne_not,
                         "Whitespace normalisation (unnormalised)",
                         __LINE__, __FUNCTION__);   
}

void tc_check_torture_syntax(lcut_tc_t *tc, void *data)
{
    const char* testxml =
         "<doc><e1   \ta1='v1'\n\ra2='v\"2' \n\n\ra3=\"v'''3\">\n\r\t\t  \"text 'text\r\nx\n\ry\n\nz</e1 \t\n ><e2 a1='v1' /><e3 \t/></doc>";
    char* expected[] = {
        "(doc",
        "Aa1 CDATA v1",
        "Aa2 CDATA v\"2",
        "Aa3 CDATA v'''3",
        "(e1",
        // See XML 1.1 spec: http://www.w3.org/TR/xml11/#sec-line-ends
        "- \"text 'text x y z",
        //"-\\n\\n\\t\\t  \"text 'text\\nx\\n\\ny\\n\\nz",
        ")e1",
        "Aa1 CDATA v1",
        "(e2",
        ")e2",
        "(e3",
        ")e3",
        ")doc",
    };
    int ne = sizeof(expected)/sizeof(expected[0]);

    assert_strings_match(tc, testxml, 0, expected, ne,
                         "Torture syntax",
                         __LINE__, __FUNCTION__);
}

void tc_check_digest(lcut_tc_t *tc, void *data)
{
    InputSource* src;
    ParserState* parser;

    // openssl sha1 ../../../test/testcases/misc/cdata.norm
    const char correct[] = "f0823b9d45dd32e4fa97efe5455ba83b6b2771e1";

#define SHA1_LENGTH 20
    char actual[SHA1_LENGTH*2+1];

    if (xoxa_digest_library_ident() == NULL) {
        // that's fine -- don't object to this, since presumably the
        // digest functionality wasn't compiled in
    } else {
        byte* digest = NULL;
        int i;
        size_t digest_len;
        const char* filetestdir = getenv("FILETESTS");
        char cdata_file[PATH_MAX];
        int status;
        

        snprintf(cdata_file, PATH_MAX, "%s/%s",
                 (filetestdir == NULL ? "../../test/testcases/misc" : filetestdir),
                 "cdata.xml");
        src = xoxa_new_input_source_from_filename(cdata_file);
        FAIL_UNLESS_LCUT_ASSERT(tc, src != NULL,
                                "Can't find test file: %s", cdata_file);

        parser = xoxa_new_parser(src);

        // digest_ctx = xoxa_digest_initialize(parser, "SHA1");
        // xoxa_parse(parser);
        // digest = xoxa_digest_finalize(parser, digest_ctx, &digest_len);
        status = xoxa_parse_for_digest(parser, "SHA1");

        FAIL_UNLESS_LCUT_ASSERT(tc, (status == 0),
                                "xoxa_parse_for_digest FAILED: %s\n", xoxa_error_message(parser));

        digest = xoxa_get_digest(parser, &digest_len);
        LCUT_INT_EQUAL(tc, digest_len, SHA1_LENGTH);
        for (i=0; i<digest_len; i++) {
            sprintf(&actual[2*i], "%02x", digest[i]);
        }
        LCUT_STR_EQUAL(tc, actual, correct);

        free(digest);
    }
#undef SHA1_LENGTH
}

void tc_check_gpg_parse_status(lcut_tc_t* tc, void* data)
{
    Buf* response_good
        = bufinit_with((byte*)
                       "[GNUPG:] SIG_ID lZqRXRGXbSynoiHthIYwxIf4lts 2013-10-09 1381337543\n"
                       "[GNUPG:] GOODSIG 719004BF4E6C1920 Xoxa Tester (passphrase 'abc') <none@example.org>\n"
                       "[GNUPG:] VALIDSIG 8D51C135C1E2730C42F4CBCE719004BF4E6C1920 2013-10-09 1381337543 0 4 0 17 2 00 8D51C135C1E2730C42F4CBCE719004BF4E6C1920\n"
                       "[GNUPG:] TRUST_ULTIMATE\n");
    Buf* response_bad
        = bufinit_with((byte*)
                       "[GNUPG:] BADSIG 719004BF4E6C1920 Xoxa Tester (passphrase 'abc') <none@example.org>\n");
    SignatureDetail sd;
    Channel* ch;


    ch = new_channel(CHANNEL_READ);
    // hack in the above buffers in turn
    buffree(ch->buf);
    ch->buf = response_good;

    LCUT_ASSERT(tc, "gpg good output parsed", gpg_parse_status(ch, &sd) != NULL);
    LCUT_ASSERT(tc, "gpg good sig", sd.status == good_sig);
    LCUT_STR_EQUAL(tc,
                   "Xoxa Tester (passphrase 'abc') <none@example.org>",
                   sd.detail.goodsig.username);

    buffree(ch->buf);
    ch->buf = response_bad;

    LCUT_ASSERT(tc, "gpg bad output parsed", gpg_parse_status(ch, &sd) != NULL);
    LCUT_ASSERT(tc, "gpg bad sig", sd.status == bad_sig);
    LCUT_STR_EQUAL(tc,
                   "Xoxa Tester (passphrase 'abc') <none@example.org>",
                   sd.detail.badsig.username);

    channel_free(ch);
}

// The following function checks _only_ the pgp_verify_detached
// function, with already-separated content and signature; it does not
// test the mechanics of extracting the content and signature from an
// XML file.
void tc_check_gpg_verifications(lcut_tc_t* tc, void* data)
{
    struct {
        const char* legend;
        const char* msg;
        const byte* sig;
        int process_status;
        GPGResult verify_status;
        const char* good_sig_username;
    } expected[] = {
        {
            "good signature",
            "This is a text file.\n"
            "It doesn't contain anything exciting,\n"
            "but it can be signed.\n", 

            (byte*)
            "-----BEGIN PGP SIGNATURE-----\n"
            "Version: GnuPG v1\n"
            "\n"
            "iEYEABECAAYFAlTcww0ACgkQcZAEv05sGSCeUwCg9dauzKwu/NOFQ6V4rBVaBmEB\n"
            "fuoAn0wRpT6h6r+VclB4apWlcXy9Hjz8\n"
            "=2jVu\n"
            "-----END PGP SIGNATURE-----\n",

            0, good_sig, "Xoxa Tester (passphrase 'abc') <none@example.org>" 
        },
        {
            "bad signature",
            "BAD BAD BAD\n"
            "This is a text file.\n"
            "It doesn't contain anything exciting,\n"
            "but it can be signed.\n",

            (byte*)
            "-----BEGIN PGP SIGNATURE-----\n"
            "Version: GnuPG v1\n"
            "\n"
            "iEYEABECAAYFAlTcww0ACgkQcZAEv05sGSCeUwCg9dauzKwu/NOFQ6V4rBVaBmEB\n"
            "fuoAn0wRpT6h6r+VclB4apWlcXy9Hjz8\n"
            "=2jVu\n"
            "-----END PGP SIGNATURE-----\n",

            1, bad_sig, NULL 
        },
        {
            "missing signature",
            "This is a text file.\n"
            "It doesn't contain anything exciting,\n"
            "but it can be signed.\n",
            // no signature provided
            // (check this produces an error rather than a crash)
            NULL, 

            2, unknown, NULL 
        },
        {
            // this is a valid signature, but using a public key that
            // the test harness doesn't know about
            "good but unknown signature",
            "This is a text file.\n"
            "It doesn't contain anything exciting,\n"
            "but it can be signed.\n", 

            (byte*)
            "-----BEGIN PGP SIGNATURE-----\n"
            "Version: GnuPG v1\n"
            "\n"
            "iQEcBAABCgAGBQJU3gNnAAoJEGH9g7GxnshFHqwIAKKNYfvNpWsZK6r3lMC/EIWu\n"
            "PUAOjPnAvYA3D4WeiiEeYCj7ZUHkaKs7+JhFXJRrJQV6hzH32qnqZJaZO0EHrduU\n"
            "Oh9FDe1oPtIXXOjImfvUMMhGvGm4B2jmq0+dLl8oLT/417xLG3sJmbJBAVj3oGOm\n"
            "Q94qo04/KrHtf72cHmaKWjEDoBdberX3kxVJjHpCVVsTZpAaCEP9AFi0zsPNwcdI\n"
            "YcHEB9nd+yksSLs3BH7mQZRkFNUytMqRa4eFaw6jkgCOGsfjfuiXMxdyzeWTBjRq\n"
            "xASZDR2kwbmry8PTON7n8pf0+rXDi2If/ofUIP5bWFGMQZ1nalmDi3TEDybEIOQ=\n"
            "=t+z8\n"
            "-----END PGP SIGNATURE-----\n",

            2, unknown, NULL
        },
        {
            // the signature is just garbage
            "garbage signature",
            "This is a text file",
            (byte*)";lijasd;flkjas;ldkfja;lksjdf",
            2, unknown, NULL
        },
        {
            "empty input file",
            "",
            (byte*) // this is a correct test user signature for the empty file
            "-----BEGIN PGP SIGNATURE-----\n"
            "Version: GnuPG v1\n"
            "\n"
            "iEYEABECAAYFAlTeBgwACgkQcZAEv05sGSCbnACgkanlwcCAECdG1QI8adV9jQYG\n"
            "0koAoIbT8EUqSrRLq3x2aoYr2S0Oq6tc\n"
            "=THNZ\n"
            "-----END PGP SIGNATURE-----\n",
            0, good_sig, "Xoxa Tester (passphrase 'abc') <none@example.org>"
        },
    };
    int n_expected = sizeof(expected)/sizeof(expected[0]);

    int i;

#define MY_LCUT_INT_EQUAL(tc, exp, actual) \
    lcut_int_equal(tc, exp, actual, __LINE__, expected[i].legend, __FILE__)
#define MY_LCUT_STR_EQUAL(tc, exp, actual) \
    lcut_str_equal(tc, exp, actual, __LINE__, expected[i].legend, __FILE__)


    for (i=0; i<n_expected; i++) {
        Buf* sig;
        Channel* sig_ch;

        Channel* status_ch;

        ProcessManager* pm;
        SignatureDetail sd;
        int status;

        //fprintf(stderr, "check: %s\n", expected[i].legend);

        sig_ch = new_channel(CHANNEL_WRITE);
        if (expected[i].sig != NULL) {
            sig = bufinit_with(expected[i].sig);
            channel_add_buffer(sig_ch, sig);
        }

        status_ch = new_channel(CHANNEL_READ);

        pm = pgp_verify_detached(sig_ch, status_ch);

        subprocess_send(pm, expected[i].msg, strlen(expected[i].msg));
        status = subprocess_send(pm, NULL, 0);

        MY_LCUT_INT_EQUAL(tc, expected[i].process_status, status);
        if (status != expected[i].process_status) {
            fprintf(stderr, "GPG failed (sample %d); ---- stdout...\n%s\n---- stderr...\n%s\n",
                    i, bufstring(pm->ch[0]->buf), bufstring(pm->ch[1]->buf));
        }

        gpg_parse_status(status_ch, &sd);
        MY_LCUT_INT_EQUAL(tc, expected[i].verify_status, sd.status);
        if (sd.status == good_sig) {
            MY_LCUT_STR_EQUAL(tc,
                              expected[i].good_sig_username,
                              sd.detail.goodsig.username);
        }

        subprocess_free(pm);
    }

#undef MY_LCUT_INT_EQUAL
#undef MY_LCUT_STR_EQUAL
}

void tc_check_gpg_xml(lcut_tc_t* tc, void* data)
{
    // The XML test here is intended to be equivalent to the test in
    // xoxa/java/src/test/java/uk/me/nxg/xoxa/SigningXMLReaderTest.java.
    // See the comments there for discussion about how the XML file is constructed.
    const char* xml = "<doc>\n<p>\n\nHello</p> <?signature algorithm=\"http://ns.nxg.me.uk/xoxa/algo#openpgp\" content='-----BEGIN PGP SIGNATURE-----\nVersion: GnuPG v1\n\niEYEABECAAYFAlO6lTwACgkQcZAEv05sGSBCigCghTjJzPqoj57oa7pLT326TuKd\nlx0An33aOLalsg46gBM4avmOlzTvtiGm\n=lyCi\n-----END PGP SIGNATURE-----\n'?>\n <p>\n there</p></doc>";
    InputSource* src;
    ParserState* PS;
    SignatureDetail* sd;
    GPGResult expected_result = good_sig;


    src = xoxa_new_input_source_from_string(xml);
    PS = xoxa_new_parser(src);

    sd = xoxa_parse_for_pgp_verify(PS);

    LCUT_TRUE(tc, sd != NULL);
    LCUT_INT_EQUAL(tc, expected_result, sd->status);
    if (sd->status == good_sig) {
        // ... as expected
        LCUT_STR_EQUAL(tc,
                       "Xoxa Tester (passphrase 'abc') <none@example.org>",
                       sd->detail.goodsig.username);
    }
}



int main() 
{
    lcut_ts_t * suite = NULL;

    LCUT_TEST_BEGIN("Xoxa tests", NULL, NULL);

    LCUT_TS_INIT(suite, "Xoxa tests", NULL, NULL);
    LCUT_TC_ADD(suite, "Simple parse", tc_check_simple_parse, NULL, NULL, NULL);
    LCUT_TC_ADD(suite, "Entity expansion", tc_check_entity_expansion,
               NULL, NULL, NULL);
    LCUT_TC_ADD(suite, "Line-end normalisation", tc_check_line_end_normalisation,
               NULL, NULL, NULL);
    LCUT_TC_ADD(suite, "Whitespace normalisation",
               tc_check_whitespace_normalisation,
               NULL, NULL, NULL);
    LCUT_TC_ADD(suite, "Torture syntax", tc_check_torture_syntax, NULL, NULL, NULL);
    LCUT_TC_ADD(suite, "Error parse", tc_check_error_parse, NULL, NULL, NULL);
    LCUT_TC_ADD(suite, "Digests", tc_check_digest, NULL, NULL, NULL);
    LCUT_TC_ADD(suite, "GPG parse status", tc_check_gpg_parse_status, NULL, NULL, NULL);
    LCUT_TC_ADD(suite, "GPG sig verifications", tc_check_gpg_verifications, NULL, NULL, NULL);
    LCUT_TC_ADD(suite, "tc_check_gpg_xml", tc_check_gpg_xml, NULL, NULL, NULL);
    LCUT_TS_ADD(suite);

    LCUT_TEST_RUN();
    LCUT_TEST_REPORT();
    LCUT_TEST_END();

    LCUT_TEST_RESULT();
}
