/* This file is part of the Xoxa library: https://bitbucket.org/nxg/xoxa
 * Copyright 2013, 2014, Norman Gray <http://nxg.me.uk>.
 * It is distributed under the terms of the BSD 2-clause Licence;
 * see the file LICENCE for details.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <assert.h>
#include <ctype.h>

#define XOXA_PRIVATE_INTERFACE 1
#include "util.h"
#undef XOXA_PRIVATE_INTERFACE
#include "config.h"
#include "ident.h"

byte* copystring(const byte* s)
{
    byte* ns;
    const byte* p;
    size_t len = 0;

    for (p=s; *p!=0; p++) {
        len += sizeof(byte);
    }
    len++;                      // include '\0'
    if ((ns = (byte*)malloc(len)) == NULL) {
        fatal("Can't allocate space for string <%s>\n", s);
    }
    return memcpy(ns, s, len);
}

#define BUFOK(B) assert(B->buf != NULL && B->nwritten < B->len)

static Buf* bufinit_(size_t init_length)
{
    Buf* b;
    if ((b = (Buf*)malloc(sizeof(Buf))) == NULL) {
        fatal("no space to allocate Buf");
    }

    b->len = init_length;
    if ((b->buf = (byte*)malloc(b->len)) == NULL) {
        fatal("can't allocate %ld bytes for Buf", b->len);
    }
    b->nwritten = 0;

    return b;
}

// Support writing formatted text to a fixed-size buffer.
Buf* bufinit(void)
{
    Buf* b = bufinit_(32);
    BUFOK(b);
    return b;
}

Buf* bufinit_with(const byte* init_string)
{
    size_t len = strlen((const char*)init_string);
    Buf* b = bufinit_(len + 1);
    bufstrncpy(b, init_string, len);
    BUFOK(b);
    return b;
}

void bufreset(Buf* B)
{
    B->nwritten = 0;
}

size_t buflength(Buf* B)
{
    return B->nwritten;
}

// Remove trailing whitespace
Buf* bufcrop(Buf* B)
{
    while (B->nwritten > 1 && isspace(B->buf[B->nwritten - 1])) {
        B->nwritten--;
    }

    // now deal with the very much less common cases of lines ending
    // with U+0085 or U+2028.  These don't have to be at all efficient
    if (B->nwritten >= 2 && isxmlspace(&B->buf[B->nwritten - 2]) == 2) {
        B->nwritten -= 2;
        bufcrop(B);             // recurse
    }
    if (B->nwritten >= 3 && isxmlspace(&B->buf[B->nwritten - 3]) == 3) {
        B->nwritten -= 3;
        bufcrop(B);             // recurse
    }

    BUFOK(B);
    return B;
}

// ensure that there are at least n more bytes available in the buffer
// and that BUFOK remains true
static Buf* buf_ensure_space_(Buf* B, size_t n)
{
    if (B->len - B->nwritten < (n+1)) {
        while (B->len - B->nwritten < (n+1)) {
            B->len *= 3;
            B->len /= 2;
        }
        if ((B->buf = (byte*)realloc((void*)B->buf, B->len)) == NULL) {
            fatal("unable to reallocate Buf space to %ld", B->len);
        }
    }
    // the following means that BUFOK will remain true
    // even after n bytes are added
    assert(B->len - B->nwritten >= (n+1));

    return B;
}

// return the assembled string -- the caller must free this
byte* bufstring_copy(Buf* B)
{
    assert(B != NULL);
    BUFOK(B);                   // guarantees nwritten < len

    B->buf[B->nwritten] = '\0';
    return copystring(B->buf);
}

// return the assembled string -- the caller must NOT free this
byte* bufstring(Buf* B)
{
    assert(B != NULL);
    BUFOK(B);                   // guarantees nwritten < len

    B->buf[B->nwritten] = '\0';
    return B->buf;
}

void buffree(Buf* b)
{
    if (b == NULL)
        return;

    assert(b->buf != NULL);
    free(b->buf);
    free(b);
}

Buf* bufprintf(Buf* B, const char* fmt, ...)
{
    size_t nw;
    size_t space_available;
    va_list ap;

    BUFOK(B);

    va_start(ap, fmt);
    space_available = B->len - B->nwritten;
    nw = vsnprintf((char*)(B->buf + B->nwritten),
                   space_available,
                   fmt,
                   ap);
    // nw is the number of characters that would have been written if
    // space_available were unlimited, so if nw > space_available, we
    // didn't write everything we wanted to, so we'll need to extend
    // the buffer.
    //
    // Note nw does not include the terminating '\0'; therefore we have to
    // check that (nw+1) is entirely accommodated within space_available

    if (! (nw+1 <= space_available)) {
        va_end(ap);
        va_start(ap, fmt);      /* reset ap */
        buf_ensure_space_(B, nw+1);

        space_available = B->len - B->nwritten;
        nw = vsnprintf((char*)(B->buf + B->nwritten),
                       space_available,
                       fmt,
                       ap);
        assert(nw+1 <= space_available);
    }

    B->nwritten += nw;

    va_end(ap);

    // the count nwritten does not include the terminating '\0', which
    // is why the invariant for this function requires that nwritten
    // is shorter than len (there's always space for buf[nwritten] = '\0')
    BUFOK(B);

    return B;
}

Buf* bufchar(Buf* B, const byte c)
{
    BUFOK(B);

    buf_ensure_space_(B, 1);
    B->buf[B->nwritten] = c;
    B->nwritten++;

    BUFOK(B);

    return B;
}

// bstrcpy is like strcpy, except with bytes
Buf* bufstrcpy(Buf* dst, const byte* src)
{
    const byte* p;
    size_t len = 0;

    for (p=src; *p != 0; p++) {
        len++;
    }

    return bufstrncpy(dst, src, len);
}

Buf* bufstrncpy(Buf* dst, const byte* src, const size_t len)
{
    if (len > 0) {
        buf_ensure_space_(dst, len);
        memcpy(dst->buf+dst->nwritten, src, len);
        dst->nwritten += len;
        BUFOK(dst);
    }
    return dst;
}

int bufstrcmp(const byte* b1, const byte* b2)
{
    while (1) {
        if (*b1 == *b2) {
            if (*b1 == 0) {
                return 0;
            } else {
                b1++;
                b2++;
            }
        } else {
            return *b1-*b2;
        }
    }
}


// print an error message and exit
void fatal(const char* fmt, ...)
{
    va_list ap;

    va_start(ap, fmt);

    fputs("libxoxa (fatal): ", stderr);
    vfprintf(stderr, fmt, ap);
    fputs("\n", stderr);

    va_end(ap);

    exit(1);
}

// print a suitably-formatted warning message, but don't exit
void warning(const char* fmt, ...)
{
    va_list ap;

    va_start(ap, fmt);

    fputs("libxoxa: ", stderr);
    vfprintf(stderr, fmt, ap);
    fputs("\n", stderr);

    va_end(ap);

    return;
}

// format a new error message string, and set the
// ParserState->error_message to this.
// If abandon_parse_p is non-zero, then we additionally call xoxa_abandon_parse_.
void xoxa_set_error_message_(ParserState* PS,
                             int abandon_parse_p,
                             const char* fmt, ...)
{
    va_list ap;
    int msglen = 256; // more than necessary -- doesn't really matter if it's cropped
    char* buf;

    if ((buf = malloc(msglen)) == NULL) {
        fatal("unable to allocate %d bytes for error message!", msglen);
    }

    // if (PS->error_message != NULL) {
    //     if (SHOWING_MESSAGES(PS)) {
    //         fprintf(stderr, "xoxa_util: discarding previous error message (%s)\n", PS->error_message);
    //     }
    //     free(PS->error_message);
    // }

    va_start(ap, fmt);
    vsnprintf(buf, msglen, fmt, ap);
    va_end(ap);

    if (PS->error_message == NULL) {
        PS->error_message = buf;
    } else {
        size_t newlen = strlen(PS->error_message)
            + strlen(buf)
            + strlen(" []")
            + 1;
         char* newbuf = malloc(newlen);

        if (newbuf == NULL) {
            fatal("Can't allocate memory for extended error message");
        }

        snprintf(newbuf, newlen, "%s [%s]", PS->error_message, buf);
        free(PS->error_message);
        free(buf);
        PS->error_message = newbuf;
    }

    if (abandon_parse_p) {
        xoxa_abandon_parse_(PS);
    }
}

// Called within a parse, abandon it, and do a longjmp back to the
// calling point. This should never be called with a null P, or a null
// P->error_jmp_buf, so class those as fatal errors.
void xoxa_abandon_parse_(ParserState* P)
{
    if (P == NULL || P->error_jmp_buf == NULL) {
        if (P->error_message != NULL) {
            fprintf(stderr, "err: %s\n", P->error_message);
        }
        fatal("abandon_parse called with null %s",
              (P == NULL ? "P" : "error_jmp_buf"));
    }

    longjmp(*P->error_jmp_buf, 1);
}

void xoxa_jmp_set_(ParserState* P, jmp_buf* env)
{
    if (P == NULL) {
        fatal("xoxa_jmp_set_: bad call -- P NULL");
    } else if (P->error_jmp_buf == NULL) {
        P->error_jmp_buf = env;
    } else {
        fatal("xoxa_jmp_set_: bad call -- error_jmp_buf was not NULL");
    }
}

void xoxa_jmp_clear_(ParserState* P)
{
    if (P == NULL) {
        fatal("xoxa_jmp_clear_: bad call -- P NULL");
    } else {
        P->error_jmp_buf = NULL;
    }
}

// Return non-zero if the character pointed to by p is an XML space.
// We also detect characters U+0085 and U+2028, which are newline
// characters in XML 1.1, by spotting their UTF-8 encoding.
// Return the number of bytes to skip (that is, 1 for the
// 'traditional' whitespace characters, and 2 & 3 respectively for NEL
// and U+2028.
int isxmlspace(const byte* p)
{
    int rval = 0;               // FALSE

    if (p[0] == '\0') {
        // mark this as non-space, so that we stop skipping at EOS
        // when we're skipping 'spaces'
        rval = 0;
    } else if (p[0] <= 0x20) {
        // White space: the only whitespace characters we expect are
        // those described in XML 1.1 Sect. 2.3, namely 0x20, 0x9, 0xa
        // and 0xd.  Other characters below 0x20 are illegal (in XML
        // 1.0, Sect. 2.2) or 'discouraged' (in XML 1.1), and so
        // should not appear at all.  Legal or not, we treat all of
        // these as 'whitespace'.
        rval = 1;
    } else if (p[0] == 0xc2 && p[1] == 0x85) {
        // U+0085, NEL
        rval = 2;
    } else if (p[0] == 0xe2 && p[1] == 0x80 && p[2] == 0xa8) {
        // U+2028
        rval = 3;
    } else {
        rval = 0;
    }

    return rval;
}

const char* xoxa_version_string(void)
{
#if _PACKAGE_RELEASED
    return PACKAGE_VERSION ", released " PACKAGE_RELEASEDATE;
#else
    return PACKAGE_VERSION " (build " IDENT_STRING ", " IDENT_DATE ")";
#endif
}
