/* This file is part of the Xoxa library: https://bitbucket.org/nxg/xoxa
 * Copyright 2013--2015, Norman Gray <http://nxg.me.uk>.
 * It is distributed under the terms of the BSD 2-clause Licence;
 * see the file LICENCE for details.
 */
#define XOXA_PRIVATE_INTERFACE 1
#include "xml_expat.h"
#include "util.h"
#undef XOXA_PRIVATE_INTERFACE
#include "pgp.h"

#include "digest.h"

#include <stdio.h>
#include <fcntl.h>
#include <string.h>
#include <assert.h>
#include <unistd.h>
#include <ctype.h>

static int output_string_with_whitespace_skipping(Buf* buf,
                                                  const XML_Char* c,
                                                  size_t len,
                                                  char *skipping_whitespace);
static void write_out_accumulated_text(ParserState* P);
static int xoxa_parse_file(ParserState* P);

static void xoxa_maybe_pop_output_callback_(ParserState* P);
static void xoxa_call_callback_(ParserState* P, byte* buf, size_t len);

// flags for ParserState->text_skipping_whitespace
#define SKIPPING_WHITESPACE_COLLAPSE 0
#define SKIPPING_WHITESPACE_TRUE 1
#define SKIPPING_WHITESPACE_NONORM 2



/**
 * Creates an InputSource from the named file.
 * The InputSource should be handed to a {@link xoxa_new_parser} function,
 * which will free it when necessary.
 * Returns NULL if there's any problem.
 * (in fact, the only problem is memory exhaustion, which causes an
 * immediate exit)
 * @param filename the file which is to be read in
 * @return a new InputSource
 */
InputSource* xoxa_new_input_source_from_filename(const char* filename)
{
    InputSource* src;
    FILE* file;

    MALLOC(src, InputSource);

    if ((file = fopen(filename, "r")) == NULL) {
        src = NULL;
    } else {
        src->type = INPUT_SOURCE_FILE;
        src->basename = filename;
        src->in.f = file;
    }

    return src;
}

/**
 * Creates an InputSource from the named file descriptor.
 * The InputSource should be handed to a {@link xoxa_new_parser} function,
 * which will free it when necessary.
 * Returns NULL if there's any problem.
 * (in fact, the only problem is memory exhaustion, which causes an
 * immediate exit)
 * @param fd the identifier of the file descriptor which is to be read in
 * @return a new InputSource
 */
InputSource* xoxa_new_input_source_from_fd(const int fd)
{
    InputSource* src;

    MALLOC(src, InputSource);

    src->type = INPUT_SOURCE_FILE;
    src->basename = NULL;
    src->in.f = fdopen(fd, "r");

    return src;
}

/**
 * Creates an InputSource from the given content, which should be an
 * string containing XML.  This string is not copied, so should not be freed
 * during a parse.
 * The InputSource should be handed to a {@link xoxa_new_parser} function,
 * which will free it when necessary.
 * Returns NULL if there's any problem.
 * (in fact, the only problem is memory exhaustion, which causes an
 * immediate exit)
 * @param xmlcontent the string which is to be read in
 * @return a new InputSource
 */
InputSource* xoxa_new_input_source_from_string(const char* xmlcontent)
{
    InputSource* src;

    MALLOC(src, InputSource);

    src->type = INPUT_SOURCE_STRING;
    src->basename = NULL;
    src->in.buffer = xmlcontent;

    return src;
}

// Separator of namespace and localname used by Expat
#define NS_SEPARATOR '\xFF'

// comparison function for 0-terminated byte strings
static int bstrcmp(const void* p1, const void* p2)
{
    const byte* b1 = *((byte**)p1);
    const byte* b2 = *((byte**)p2);

    while (1) {
        if (*b1 == 0) {
            if (*b2 == 0) return 0; // equal
            return -1;
        } else if (*b2 == 0) {
            return +1;
        } else if (*b1 == *b2) {
            b1++;
            b2++;
        } else if (*b1 > *b2) {
            return +1;
        } else {
            return -1;
        }
    }
}

// Expat handler functions

static void expat_element_start(void *data, const XML_Char *el, const XML_Char **attr)
{
    int i;
    ParserState* P = (ParserState*)data;
    char* nsp;
    byte* output;
#define MAXATT 32
    // make this statically allocated for the moment
    // (not pretty, but 32 is a lot of attributes...)
    size_t fmtatt[MAXATT];
    byte* fmtatt_b[MAXATT];
    int natt;
    char skipping_whitespace_init;


    assert(P != NULL);
    assert(P->B != NULL);

    P->element_depth++;

    write_out_accumulated_text(P);
    bufreset(P->B);

    // if there are other callbacks, call them now
    if (P->extra_StartElementHandler) {
        (*P->extra_StartElementHandler)(P->extra_UserData, el, attr);
    }

    if (P->flags.report_all) {
        skipping_whitespace_init = SKIPPING_WHITESPACE_NONORM;
    } else {
        skipping_whitespace_init = SKIPPING_WHITESPACE_COLLAPSE;
    }

    // build up an array of attribute strings
    natt = 0;
    for (i=0; attr[i]; i+=2) {
        // skipping_whitespace is potentially modified in
        // output_string_with_whitespace_skipping
        char skipping_whitespace;

        if (natt == MAXATT) {
            fatal("Unexpectedly large number (>%d) of attributes on element %s", MAXATT, el);
        }

        fmtatt[natt] = buflength(P->B); // use this as an offset into the buffer
        nsp = strchr(attr[i], NS_SEPARATOR);

        skipping_whitespace = skipping_whitespace_init;

        if (nsp == NULL) {
            bufprintf(P->B, "A%s CDATA ", attr[i]);
            output_string_with_whitespace_skipping(P->B,
                                                   attr[i+1],
                                                   0,
                                                   &skipping_whitespace);
            bufprintf(P->B, "%c", NS_SEPARATOR);
        } else {
            bufprintf(P->B, "B%.*s %s CDATA ", nsp-attr[i], attr[i], nsp+1);
            output_string_with_whitespace_skipping(P->B,
                                                   attr[i+1],
                                                   0,
                                                   &skipping_whitespace);
            bufprintf(P->B, "%c", NS_SEPARATOR);
        }
        natt++;
    }

    // sort these...
    output = bufstring_copy(P->B);
    // first, split this output at NS_SEPARATOR
    for (i=0; i<natt; i++) {
        byte* p;
        fmtatt_b[i] = output+fmtatt[i];
        for (p=fmtatt_b[i]; *p!=(byte)NS_SEPARATOR; p++) {
            // nothing
        }
        *p = '\0';
    }
    qsort((void*)fmtatt_b, natt, sizeof(fmtatt_b[0]), &bstrcmp);

    for (i=0; i<natt; i++) {
        byte crlf[2] = {'\r', '\n'};

        xoxa_call_callback_(P, fmtatt_b[i], strlen((const char*)fmtatt_b[i]));
        xoxa_call_callback_(P, crlf, 2);
    }
    free(output);

    bufreset(P->B);

    nsp = strchr(el, NS_SEPARATOR);
    if (nsp == NULL) {
        // no namespace
        P->B = bufprintf(P->B, "(%s\r\n", el);
    } else {
        P->B = bufprintf(P->B, "[%.*s %s\r\n", nsp-el, el, nsp+1);
    }

    xoxa_call_callback_(P, bufstring(P->B), buflength(P->B));

    // Set the parser's skipping-whitespace flag to the correct
    // initial value for processing the text content of the element.
    P->text_skipping_whitespace = skipping_whitespace_init;
#undef MAXATT
}

static void expat_element_end(void *data, const XML_Char *el)
{
    ParserState* P = (ParserState*)data;
    char* nsp;


    assert(P != NULL);
    assert(P->B != NULL);

    write_out_accumulated_text(P);
    bufreset(P->B);

    if (P->extra_EndElementHandler) {
        (*P->extra_EndElementHandler)(P->extra_UserData, el);
    }

    nsp = strchr(el, NS_SEPARATOR);
    if (nsp == NULL) {
        // no namespace
        P->B = bufprintf(P->B, ")%s\r\n", el);
    } else {
        P->B = bufprintf(P->B, "]%.*s %s\r\n", nsp-el, el, nsp+1);
    }

    xoxa_call_callback_(P, bufstring(P->B), buflength(P->B));

    assert(P->element_depth > 0);
    P->element_depth--;

    xoxa_maybe_pop_output_callback_(P);
}

static void expat_character_data_handler(void *data,
                                         const XML_Char *s,
                                         int len)
{
    ParserState* P = (ParserState*)data;

    assert(P != NULL);

    output_string_with_whitespace_skipping(P->text,
                                           s,
                                           len,
                                           &P->text_skipping_whitespace);
    // do not call P->extra_CharacterDataHandler here, since this is called
    // in write_out_accumulated_text, below
}

static void write_out_accumulated_text(ParserState* P)
{
    size_t len;

    assert(P != NULL);
    len = buflength(P->text);

    // if len==0, there's nothing to write;
    // if (len==1 and skipping_whitespace = TRUE),
    //   then all there is to write is whitespace
    //   (and we're in the normalization mode);
    // in either case do nothing
    if (! (len == 0
           || (len == 1
               && P->text_skipping_whitespace == SKIPPING_WHITESPACE_TRUE))) {
        byte dash[] = { '-' };
        byte crlf[] = { '\r', '\n' };

        xoxa_call_callback_(P, dash, 1);
        xoxa_call_callback_(P, bufstring(P->text), buflength(P->text));
        xoxa_call_callback_(P, crlf, 2);
    }

    if (P->extra_CharacterDataHandler) {
        // XXX: I'm somewhat uncertain about this cast.  There's a check,
        // in xoxa_new_parser, that the sizeof values for byte and
        // XML_Char are the same, but are there sign and
        // unicode-related problems possible, too?
        (*P->extra_CharacterDataHandler)(P->extra_UserData,
                                         (const XML_Char*)bufstring(P->text),
                                         buflength(P->text));
    }

    bufreset(P->text);

    // Set text_skipping_whitespace to the correct value to process
    // the text content correctly.
    if (P->flags.report_all) {
        P->text_skipping_whitespace = SKIPPING_WHITESPACE_NONORM;
    } else {
        P->text_skipping_whitespace = SKIPPING_WHITESPACE_COLLAPSE;
    }
}

static void expat_namespace_start(void *data,
                                  const XML_Char *prefix,
                                  const XML_Char *uri)
{
    ParserState* P = (ParserState*)data;

    assert(P != NULL);
    write_out_accumulated_text(P);

    if (P->extra_StartNamespaceDeclHandler) {
        (*P->extra_StartNamespaceDeclHandler)(P->extra_UserData, prefix, uri);
    }

    if (P->flags.report_all) {
        bufreset(P->B);
        bufprintf(P->B, "M%s %s\r\n", (prefix ? prefix : ""), uri);
        xoxa_call_callback_(P, bufstring(P->B), buflength(P->B));
    }
}

static void expat_namespace_end(void *data,
                                const XML_Char *prefix)
{
    ParserState* P = (ParserState*)data;

    assert(P != NULL);
    write_out_accumulated_text(P);

    if (P->extra_StartNamespaceDeclHandler) {
        (*P->extra_EndNamespaceDeclHandler)(P->extra_UserData, prefix);
    }

    if (P->flags.report_all) {
        bufreset(P->B);
        bufprintf(P->B, "m%s\r\n", (prefix ? prefix : ""));
        xoxa_call_callback_(P, bufstring(P->B), buflength(P->B));
    }
}

static void expat_pi(void *data,
                     const XML_Char *target,
                     const XML_Char *pi_data)
{
    ParserState* P = (ParserState*)data;
    int show_it = 1;            // do we write this PI out?
    int verbose = 1;
    const char* signature_xpath = NULL;


    assert(P != NULL);

    if (!SHOWING_MESSAGES(P)) {
        verbose = 0;
    } else if (SHOWING_CHATTER(P)) {
        verbose = 2;
    }

#define ERR_PFX "libxoxa: malformed <?signature?>: "

    if (strcmp(target, "signature") == 0) {
        const char* p = pi_data;
        char quotechar;

        while (1) {
            const char* k;
            unsigned int klen;
            const char* v;
            size_t vlen;
            const char** store = NULL;

            while(isspace(*p))
                p++;
            if (*p == '\0')
                break;          // end of attributes: JUMP OUT

            k = p;
            while (isalnum(*p))
                p++;
            klen = p - k;

            if (*p != '=') {
                xoxa_set_error_message_(P, 1, "malformed attributes [%s]", pi_data);
                // bails out
            }
            p++;
            switch (*p) {
              case '\'': quotechar = '\''; break;
              case '"':  quotechar = '"'; break;
              default:
                xoxa_set_error_message_(P, 1, "malformed attributes [%s]", pi_data);
                // bails out
            }
            p++;

            v = p;
            while (*p != quotechar) {
                if (*p == '\0') {
                    xoxa_set_error_message_(P, 1,
                                            "unterminated signature attribute [%s]",
                                            pi_data);
                }
                p++;
            }
            vlen = p - v;
            p++;                // now pointing after end-quote

            if (strncmp(k, "content", klen) == 0) {
                store = &P->signature_armor;
                if (P->signature_callback != NULL) {
                    (*P->signature_callback)((const byte*)v,
                                             vlen,
                                             P->signature_callback_data);
                }
            } else if (strncmp(k, "algorithm", klen) == 0) {
                store = &P->signature_algorithm;
            } else if (strncmp(k, "target", klen) == 0) {
                store = &signature_xpath;
            } else {
                if (verbose) {
                    fprintf(stderr, ERR_PFX "unrecognised attribute %.*s (ignored)\n",
                            klen, k);
                }
                store = NULL;
            }

            if (store != NULL) {
                char* buf = malloc(vlen + 1);
                if (buf == NULL) {
                    fatal("can't malloc memory for signature attribute value");
                }
                strncpy(buf, v, vlen);
                buf[vlen] = '\0';
                *store = buf;
            }
        }

        if (P->flags.report_all) {
            show_it = 1;
        } else {
            show_it = 0;
        }
    }

    if (signature_xpath != NULL) {
        // an xpath (a target attribute) was specified
        P->signature_xpath = signature_xpath; // record this

        if (strcmp(signature_xpath, "/") == 0) {
            // This signature is for the whole document.
            // Here, P->d.digest.ctx may or may not be NULL.  If it's
            // not NULL, then we're presumably calculating a signature
            // for the whole document; if it is, then we appear not to
            // be (perhaps we're just parsing the contents of the
            // document), but that's OK
        } else if (strcmp(signature_xpath, "following::*[1]") == 0) {
            if (P->d.digest.ctx == NULL) {
                if (P->signature_algorithm == NULL) {
                    xoxa_set_error_message_(P, 1,
                                            ERR_PFX "found target attribute, but no standard attribute");
                }
                if (verbose > 1) {
                    fprintf(stderr, "Starting digest/verification of %s with %s\n",
                            signature_xpath, P->signature_algorithm);
                }
                if (strcmp(P->signature_algorithm, PGP_ALGORITHM_NAME) == 0) {
                    xoxa_pgp_verify_initialize(P);
                } else {
                    xoxa_digest_initialize(P, P->signature_algorithm);
                    // this will put something into P->d.digest.ctx
                }
            } else {
                fprintf(stderr, ERR_PFX "found target=following::*[1], but digest already in progress (ignored)\n");
            }
        } else {
            xoxa_set_error_message_(P, 1,
                                    ERR_PFX "unexpected target: %s (expecting '/' or 'following::*[1]')",
                                    P->signature_xpath);
        }
    }

    if (show_it) {
        char skipping_whitespace = SKIPPING_WHITESPACE_COLLAPSE;
        int found_chars;

        write_out_accumulated_text(P);

        bufreset(P->B);
        bufprintf(P->B, "?%s ", target);

        found_chars = output_string_with_whitespace_skipping(P->B,
                                                             pi_data,
                                                             0,
                                                             &skipping_whitespace);
        bufcrop(P->B);
        if (! found_chars) {
            // the preceding crop will have removed _all_ trailing
            // characters, including the one following the "?%s ".
            // So put it back.
            bufchar(P->B, ' ');
        }

        bufprintf(P->B, "\r\n");
        xoxa_call_callback_(P, bufstring(P->B), buflength(P->B));
    }

    if (P->extra_ProcessingInstructionHandler) {
        (*P->extra_ProcessingInstructionHandler)(P->extra_UserData, target, pi_data);
    }
#undef ERR_PFX
}

static void expat_xmldecl(void *data,
                          const XML_Char* version,
                          const XML_Char* encoding,
                          int standalone)
{
    ParserState* P = (ParserState*)data;

    assert(P != NULL);

    if (strcmp(version, "1.0") != 0) {
        if (SHOWING_MESSAGES(P)) {
            warning("document is <?xml version='%s'?>, but xoxa/expat can in principle handle only XML 1.0; there may be subtle whitespace normalization errors.", version);
            // The errors in question are to do with handling NEL
            // (U+0085) and line-separator (U+2028); cf XML 1.1 Sect. 1.3.
        }
    }

    if (P->extra_XmlDeclHandler) {
        (*P->extra_XmlDeclHandler)(P->extra_UserData, version, encoding, standalone);
    }
}

static int expat_external_entity_ref_handler(XML_Parser main_parser_state_ptr,
                                             const XML_Char *context,
                                             const XML_Char *document_base_filename,
                                             const XML_Char *systemId,
                                             const XML_Char *publicId)
{
    // Note: xoxa_new_parser calls XML_SetExternalEntityRefHandlerArg
    // with the new ParserState object, so that, despite its
    // declaration here, the main_parser_state_ptr really _is_ a
    // ParserState* pointer.  See the notes about this in the
    // description of XML_SetExternalEntityRefHandlerArg in the
    // Expat docs.
    ParserState* PS = (ParserState*)main_parser_state_ptr;
    XML_Parser main_parser = PS->parser;
    InputSource* entity_src;
    int old_cwd = -1;
    const char* bases[2];
    int base_idx;
    int error_p = 0;

    OK_PARSERSTATE(PS);

    // We _don't_ call any extra ExternalEntityRefHandler here, but
    // simply arrange that the original callbacks are called as
    // usual.  This is so that there's no opportunity for the caller
    // which lodged such a callback to read the entity.  We want to
    // read it, and we don't want the caller to get in the way.

    if (systemId == NULL) {
        // I think this should be impossible, but it's not a Xoxa
        // error, so fail with an error message
        fatal("Unexpected external entity with empty system id at line %lu",
              XML_GetCurrentLineNumber(main_parser));
    }

    // Assume, at least for the present, that the systemId and base
    // point to a file, rather than a general URI.
    bases[1] = document_base_filename;
    if (PS->src == NULL) {
        bases[0] = NULL;
    } else {
        bases[0] = PS->src->basename; // filename of the current document
    }
    base_idx = 1;                 // index of last element of bases[]

    while (base_idx >= 0) {
        if (bases[base_idx] == NULL) {
            base_idx--;         // go round again
        } else {
            const char* base = bases[base_idx];
            const char *slash = strrchr(base, '/');
            if (slash == NULL) {
                // the base URI is just a file -- nothing to do
            } else {
                // we've found the last slash in the string, so this
                // indicates the directory of the base file
                size_t dirlen = slash - base;
                int chdir_status;
                char* new_cwd = malloc(dirlen + 1);

                if (new_cwd == NULL) {
                    fatal("Can't allocate space for base-URI directory, %s (%lu bytes)",
                          base, dirlen);
                }
                memcpy(new_cwd, base, dirlen);
                new_cwd[dirlen] = '\0';

                // open the current working directory, and save its (nonzero) FD
                old_cwd = open(".", O_RDONLY);

                chdir_status = chdir(new_cwd);

                if (chdir_status != 0) {
                    // ooops
                    xoxa_set_error_message_(PS, 0,
                                            "Can't cwd to directory %s (base=%s)",
                                            new_cwd, base);
                    error_p = 1;
                }
                free(new_cwd);
                if (error_p) {
                    goto clean_up; // JUMP OUT
                }
            }
            base_idx = -1;      // exit loop
        }
    }

    entity_src = xoxa_new_input_source_from_filename(systemId);
    if (entity_src == NULL) {
        char* cwd = getcwd(NULL, 0);
        xoxa_set_error_message_(PS, 0,
                                "Can't open entity %s to read (cwd=%s)",
                                systemId, cwd);
        free(cwd);
        goto clean_up;
    }

    {
        XML_Parser sub_parser;
        ParserState* sub_PS;
        int sub_error_p = 0;

        sub_parser = XML_ExternalEntityParserCreate(main_parser, context, NULL);
        sub_PS = xoxa_make_new_parserstate_(entity_src, sub_parser, PS);
        XML_SetExternalEntityRefHandlerArg(sub_parser, sub_PS);

        sub_error_p = xoxa_parse_file(sub_PS);
        if (sub_error_p) {
            xoxa_set_error_message_(PS, 0, xoxa_error_message(sub_PS));
            error_p = 1;
        }

        xoxa_release_parser(sub_PS);     // also frees sub_parser and entity_src
    }

 clean_up:
    if (old_cwd >= 0) {
        // change back to the original WD
        fchdir(old_cwd);
        close(old_cwd);
    }

    if (error_p) {
        if (SHOWING_MESSAGES(PS)) {
            warning("Error parsing entity %s: %s", systemId, xoxa_error_message(PS));
        }
        return XML_STATUS_ERROR;
    } else {
        return XML_STATUS_OK;
    }

#undef ERR_MSG_LEN
}

static void expat_skipped_entity_handler(void *data,
                                         const XML_Char *entityName,
                                         int is_parameter_entity)
{
    ParserState* P = (ParserState*)data;

    assert(P != NULL);

    if (SHOWING_MESSAGES(P)) {
        warning("Skipped entity found (%s%s) -- I didn't expect that",
                (is_parameter_entity ? "parameter entity: " : ""),
                entityName);
    }

    if (P->extra_SkippedEntityHandler) {
        (*P->extra_SkippedEntityHandler)(P->extra_UserData, entityName, is_parameter_entity);
    }
}

/**
 * The 'default' callback receives a line of ESIS output and writes it to stdout.
 * If 'len' is zero, then the buf is null-terminated; if not, then it
 * gives the number of bytes to be written out.
 * @param buf the buffer of output data
 * @param len the length of the buffer
 * @param data the user-supplied data blob
 */
void xoxa_default_callback(const byte* buf, size_t len, void* data)
{
    if (len == 0) {
        for (len = 0; buf[len] != '\0'; len++) {
            // nothing
        }
    }
    fwrite(buf, len, 1, stdout);
}

/**
 * The 'null' callback receives a line of ESIS output and discards it.
 * @param buf the buffer of output data
 * @param len the length of the buffer
 * @param data the user-supplied data blob
 */
void xoxa_null_callback(const byte* buf, size_t len, void* data)
{
    // nothing
}


// API functions

// xoxa_make_new_parserstate_ is internal-only
// If orig is NULL, then this is a new parserstate;
// if it's non-NULL, it's a sub-parser, and the various fields
// are filled in from the original parser.
ParserState* xoxa_make_new_parserstate_(InputSource* src,
                                        XML_Parser parser,
                                        ParserState* orig)
{
    ParserState* P;

    MALLOC(P, ParserState);

    if (orig == NULL) {
        P->parser = parser;
        P->src = src;
        P->callback = NULL;
        P->element_depth = 0;
        P->B = bufinit();
        P->text = bufinit();
        P->text_skipping_whitespace = SKIPPING_WHITESPACE_COLLAPSE;
        P->flags.report_all = 0;
        P->flags.verbosity = 1;
        P->signature_armor = NULL;
        P->signature_xpath = NULL;
        P->signature_algorithm = NULL;
        P->signature_callback = NULL;
        P->signature_callback_data = NULL;
        P->error_jmp_buf = NULL;
        P->error_message = NULL;

        P->extra_StartElementHandler = NULL;
        P->extra_EndElementHandler = NULL;
        P->extra_CharacterDataHandler = NULL;
        P->extra_ProcessingInstructionHandler = NULL;
        P->extra_StartCdataSectionHandler = NULL;
        P->extra_EndCdataSectionHandler = NULL;
        P->extra_StartNamespaceDeclHandler = NULL;
        P->extra_EndNamespaceDeclHandler = NULL;
        P->extra_XmlDeclHandler = NULL;
        P->extra_SkippedEntityHandler = NULL;
        P->extra_UserData = NULL;
    } else {
        P->parser = parser;
        // The following is not orig->src: this is a sub-parser, being
        // set up to parse the _new_ source 'src', while pausing in
        // the parsing of the original, 'P->src'.
        P->src = src;
        P->B = bufinit();
        P->text = bufinit();
        P->text_skipping_whitespace = orig->text_skipping_whitespace;
        P->flags.report_all = orig->flags.report_all;
        P->flags.verbosity = orig->flags.verbosity;
        P->flags.digest_type = XOXA_DIGEST_TYPE_NONE;
        P->signature_callback = orig->signature_callback;
        P->error_message = NULL;

        P->extra_StartElementHandler =
            orig->extra_StartElementHandler;
        P->extra_EndElementHandler =
            orig->extra_EndElementHandler;
        P->extra_CharacterDataHandler =
            orig->extra_CharacterDataHandler;
        P->extra_ProcessingInstructionHandler =
            orig->extra_ProcessingInstructionHandler;
        P->extra_StartCdataSectionHandler =
            orig->extra_StartCdataSectionHandler;
        P->extra_EndCdataSectionHandler =
            orig->extra_EndCdataSectionHandler;
        P->extra_StartNamespaceDeclHandler =
            orig->extra_StartNamespaceDeclHandler;
        P->extra_EndNamespaceDeclHandler =
            orig->extra_EndNamespaceDeclHandler;
        P->extra_XmlDeclHandler =
            orig->extra_XmlDeclHandler;
        P->extra_SkippedEntityHandler =
            orig->extra_SkippedEntityHandler;
        P->extra_UserData =
            orig->extra_UserData;
    }

    return P;
}

/**
 * Returns a freshly-initialised parser.
 * After this call, the InputSource object is owned by the Parser, and
 * will be released in xoxa_release_parser.
 *
 * If there is an error, print a message to stderr and return NULL.
 *
 * @param src an initialised InputSource object, which must not be NULL
 * @return a fresh parser
 */
ParserState* xoxa_new_parser(InputSource* src)
{
    ParserState* PS;

    // The following should always be true, unless the Expat library
    // was compiled with XML_UNICODE=1 _and_ the resulting library was
    // not renamed to libexpatw.la as the Expat build instructions
    // require.
#define BUILD_BUG_ON(condition) ((void)sizeof(char[1 - 2*!!(condition)]))
    BUILD_BUG_ON(sizeof(XML_Char) != 1);

    XML_Parser p = XML_ParserCreateNS(NULL, NS_SEPARATOR);
    if (! p) {
        fatal("Couldn't allocate memory for expat parser");
    }

    PS = xoxa_make_new_parserstate_(src, p, NULL);
    OK_PARSERSTATE(PS);

    XML_SetElementHandler(p, expat_element_start, expat_element_end);
    XML_SetCharacterDataHandler(p, expat_character_data_handler);
    XML_SetNamespaceDeclHandler(p,
                                expat_namespace_start,
                                expat_namespace_end);
    XML_SetProcessingInstructionHandler(p, expat_pi);
    XML_SetXmlDeclHandler(p, expat_xmldecl);
    XML_SetExternalEntityRefHandler(p, expat_external_entity_ref_handler);
    XML_SetExternalEntityRefHandlerArg(p, (void*)PS);
    XML_SetSkippedEntityHandler(p, expat_skipped_entity_handler);
    XML_SetUserData(p, (void*)PS);

    return PS;
}

/**
 * Sets parser flags.
 * @param P an initialised parser
 * @param PF one or more flags, or-ed together
 */
void xoxa_parser_flags(ParserState* P, ParserFlags PF)
{

    if (P) {
        if (PF & XOXA_REPORT_ALL) {
            P->flags.report_all = 1;
        }
        switch (PF & XOXA_CHATTER) {
          case XOXA_QUIET:
            P->flags.verbosity = 0;
            break;
          case XOXA_FEEDBACK:
            P->flags.verbosity = 1;
            break;
          case XOXA_CHATTER:
            P->flags.verbosity = 2;
            break;
          default:
            break;
        }
    }
    return;
}

static void signature_print_callback(const byte* content, size_t len, void* data)
{
    static int fd = -1;
    size_t nwritten = 0;

    if (fd >= 0) {
        // been here before!
        // do nothing
        return;
    } else {
        fd = *(int*)data;
        free(data);
    }

    while (nwritten < len) {
        ssize_t nw;
        nw = write(fd, &content[nwritten], len-nwritten);
        if (nw < 0) {
            warning("error writing signature");
            break;
        }
        nwritten += nw;
    }
}

/**
 * Specify the file-descriptor to which the library should write a signature block.
 * If this is provided, then when the library encounters a
 * `<?signature?>` PI, it writes the contents of the
 * `armor="..."` PI 'attribute' to the given open file-descriptor.
 * The precise behaviour of the library in this respect may change in
 * future beta versions, if and when further signature mechanisms are
 * supported.
 *
 * @param P a configured parser
 * @param fd an open file descriptor
 */
void xoxa_signature_fd(ParserState* P, int fd)
{
    if (P) {
        if (fd < 1) {
            if (SHOWING_MESSAGES(P)) {
                warning("fd (%d) must be greater than 1 -- ignored", fd);
            }
        } else {
            int* fdp = malloc(sizeof(int)); // this is freed within signature_print_callback
            *fdp = fd;
            P->signature_callback = &signature_print_callback;
            P->signature_callback_data = fdp;
        }
    }
}

/**
 * Retrieve a message digest from the parser, after a parse has completed.
 * If the parsed document did not contain a digest, this returns NULL
 * (that is, it is not an error to call both this and
 * {@link xoxa_get_signature} to detect which structures are available).
 *
 * The resulting byte array is owned by the ParserState, and should
 * not be freed by the caller.
 *
 * @param P a configured parser, after a parse has completed successfully
 * @param lenp if non-NULL, this points to a location which is filled
 * in with the length of the digest in bytes
 * @return a pointer to the message digest, or NULL if P is NULL or if
 * no digest is available
 *
 * @see xoxa_get_signature
 */
byte* xoxa_get_digest(ParserState* P, size_t* lenp)
{
    unsigned int len = 0;
    byte* rval = NULL;

    if (P != NULL
        && P->flags.digest_type == XOXA_DIGEST_TYPE_MD
        && P->d.digest.bytes != NULL) {
        rval = P->d.digest.bytes;
        len = P->d.digest.len;
    }
    if (lenp != NULL) {
        *lenp = len;
    }

    return rval;
}

/**
 * Retrieve a PGP signature from the parser.
 * If the parsed document did not contain a signature, this returns NULL
 * (that is, it is not an error to call both this and
 * {@link xoxa_get_digest} to detect which structures are
 * available).
 *
 * The returned struct is owned by the parser struct, and should not
 * be freed by the caller.
 *
 * @param P a configured parser, after a parser has completed successfully
 * @return a pointer to a SignatureDetail struct, or NULL if P is NULL
 * or if no signature is available
 * @see xoxa_get_digest
 */
SignatureDetail* xoxa_get_signature(ParserState* P)
{
    SignatureDetail* rval;

    if (P == NULL) {
        rval = NULL;
    } else if (P->flags.digest_type == XOXA_DIGEST_TYPE_PGP) {
        rval = &(P->d.pgp);
    } else {
        rval = NULL;
    }

    return rval;
}

/**
 * Releases the parser and the contained input source.
 */
void xoxa_release_parser(ParserState* P)
{
#define FREE_AND_NULLIFY(x) do {                \
        if (x != NULL) {                        \
            free((void*)x);                     \
            x = NULL;                           \
        }                                       \
    } while (0)


    if (P != NULL) {
        assert(P->parser != NULL);
        XML_ParserFree(P->parser);
        P->parser = NULL;
        FREE_AND_NULLIFY(P->src);
        FREE_AND_NULLIFY(P->error_message);
        FREE_AND_NULLIFY(P->signature_armor);
        FREE_AND_NULLIFY(P->signature_algorithm);
        FREE_AND_NULLIFY(P->signature_xpath);

        if (P->flags.digest_type == XOXA_DIGEST_TYPE_MD) {
            // see assertion in digest_osx.c / digest_openssl.c
            assert(P->d.digest.bytes != NULL);
            FREE_AND_NULLIFY(P->d.digest.bytes);
        }

        free(P);
    }
}

static int xoxa_parse_string(ParserState* P)
{
    enum XML_Status status;
    int rval;
    jmp_buf jmp_env;

    OK_PARSERSTATE(P);
    assert(P->src != NULL);
    assert(P->src->type == INPUT_SOURCE_STRING);

    if (setjmp(jmp_env) == 0) {
        // normal behaviour
        xoxa_jmp_set_(P, &jmp_env);
        status = XML_Parse(P->parser,
                           P->src->in.buffer,
                           strlen(P->src->in.buffer),
                           1);
    } else {
        // the parse was abandoned with an error
        status = XML_STATUS_ERROR;
    }
    xoxa_jmp_clear_(P);

   if (status == XML_STATUS_OK) {
       rval = 0;
    } else {
       xoxa_set_error_message_(P, 0, "Parse error at character %lu: %s\n",
                               XML_GetCurrentByteIndex(P->parser),
                               XML_ErrorString(XML_GetErrorCode(P->parser)));
       rval = 1;
   }

   return rval;
}

static int xoxa_parse_file(ParserState* P)
{
    int keep_going = 1;
    FILE* f;
    int rval = 0;
    jmp_buf jmp_env;

    OK_PARSERSTATE(P);
    assert(P->src != NULL);
    assert(P->src->type == INPUT_SOURCE_FILE);
    assert(P->src->in.f != NULL);

    f = P->src->in.f;

    if (setjmp(jmp_env) == 0) {
        // normal behaviour
        xoxa_jmp_set_(P, &jmp_env);

        while (keep_going) {
            int bytes_read;
            int is_last;

            void *buf = XML_GetBuffer(P->parser, BUFSIZ);
            if (buf == NULL) {
                fatal("Unable to allocate memory in expat GetBuffer");
            }

            bytes_read = fread(buf, 1, BUFSIZ, f);
            if (ferror(f)) {
                fatal("Error reading input XML");
            }

            if (feof(f)) {
                keep_going = 0;
                is_last = 1;
            } else {
                is_last = 0;
            }

            if (XML_ParseBuffer(P->parser, bytes_read, is_last) == XML_STATUS_ERROR) {
                const char* fn = P->src->basename;
                xoxa_set_error_message_(P, 0,
                                        "Parse error at %s:%lu: %s\n",
                                        (fn == NULL ? "<input>" : fn),
                                        XML_GetCurrentLineNumber(P->parser),
                                        XML_ErrorString(XML_GetErrorCode(P->parser)));
                rval = 1;
            }
        }
    } else {
        rval = 1;
    }
    xoxa_jmp_clear_(P);

    fclose(f);
    P->src->in.f = NULL;

    return rval;
}

/**
 * Parse the input, using the configured parser.
 * If there is an error, then there will be an error message string
 * available through xoxa_error_message(ParserState*).
 * @param P a configured parser
 * @return zero on success, and non-zero on any error
 */
int xoxa_parse(ParserState* P)
{
    int rval;

    if (P == NULL) {
        warning("Bad call: xoxa_parse received null ParserState");
        return 1;
    }
    if (! PARSERSTATE_OK_TEST_(P)) {
        warning("Bad call: mangled ParserState");
        return 1;
    }
    if (P->src == NULL) {
        warning("Bad call: xoxa_parse can't be used with null InputSource");
        return 1;
    }
    assert(P->src != NULL);

    if (P->callback == NULL) {
        // to stop this, push the xoxa_null_callback
        xoxa_push_output_callback(P, &xoxa_default_callback, NULL, "default-xoxa");
    }

    switch (P->src->type) {
      case INPUT_SOURCE_STRING:
        rval = xoxa_parse_string(P);
        break;
      case INPUT_SOURCE_FILE:
        rval = xoxa_parse_file(P);
        break;
      default:
        assert(0);
    }
    return rval;
}

/**
 * Retrieve any error message which has been collected during a parse.
 * This string is owned by the parser, and must not be freed by the caller.
 * @param P a parser, which has been used to parse an input source
 */
const char* xoxa_error_message(ParserState* P)
{
    if (P == NULL) {
        return NULL;
    } else {
        return P->error_message;
    }
}

/**
 * Push a callback onto the callback stack.
 * This may be done before a parse is started, or during a parse (for
 * example at the start of an element).  In the latter case, the
 * callback is popped at the end of the element, and any associated
 * finaliser called then.
 *
 * When a callback is called, the other callbacks below it in the
 * stack are called also.
 *
 * The default callback (which is added if no callback is provided by the user)
 * writes the normalized ESIS information to the
 * standard output.  If you wish to suppress this, then you must
 * push the callback xoxa_null_callback (which does nothing).
 *
 * @param P a parser state, which must not be NULL
 * @param cb a callback function, which must not be NULL
 * @param data a blob of data to be passed to the callback function
 * when it is called
 * @param label a name for this, which is useful for debugging
 * @see xoxa_register_output_callback_finalizer
 * @see xoxa_null_callback
 */
int xoxa_push_output_callback(ParserState* P,
                              output_callback_t cb,
                              void* data,
                              const char* label)
{
    callback_frame_t* frame;

    if (P == NULL) {
        warning("Bad call to xoxa_push_output_callback: ParserState NULL");
        return 1;
    }
    if (cb == NULL) {
        warning("Bad call to xoxa_push_output_callback: cb NULL");
        return 1;
    }

    MALLOC(frame, callback_frame_t);

    frame->f = cb;
    frame->f_end = NULL;
    frame->startdepth = P->element_depth;
    frame->data = data;
    frame->label = label;
    frame->down = P->callback;

    assert(VALID_CALLBACK_FRAME(frame));

    P->callback = frame;

    if (SHOWING_CHATTER(P)) {
        fprintf(stderr, "callback push %s: element depth %d\n",
                frame->label, frame->startdepth);
    }

    return 0;
}

/**
 * Register a finaliser to be called when the top callback on the
 * stack is popped.
 *
 * @param P a parser state
 * @param cb_end a finalizer function
 * @see xoxa_push_output_callback
 */
void xoxa_register_output_callback_finalizer(ParserState* P,
                                             output_callback_end_t cb_end)
{
    callback_frame_t* frame;

    assert(P != NULL);

    frame = P->callback;

    if (frame == NULL) {
        // a programming error, but not an assertion error
        fatal("Attempt to register callback finalizer without any callback registered");
    }

    frame->f_end = cb_end;

    return;
}

static void xoxa_maybe_pop_output_callback_(ParserState* P)
{
    callback_frame_t* frame;

    assert(P != NULL);

    frame = P->callback;

    if (frame == NULL) {
        return;
    }

    if (P->element_depth == frame->startdepth) {
        output_callback_end_t cb_end = frame->f_end;

        if (SHOWING_CHATTER(P)) {
            fprintf(stderr, "callback pop %d", frame->startdepth);
            if (cb_end == NULL) {
                fprintf(stderr, " (no cleanup)\n");
            } else {
                fprintf(stderr, "; and calling %s\n", frame->label);
            }
        }

        if (cb_end != NULL) {
            (*cb_end)(P, frame->data);
        }

        P->callback = frame->down;
        free(frame);
    }
}

// This calls the top callback in the stack, _plus_ the other
// callbacks down the stack to the bottom.  Perhaps make this configurable?
//
// It's OK for P->callback to be NULL
static void xoxa_call_callback_(ParserState* P, byte* buf, size_t len)
{
    callback_frame_t* frame;

    assert(P != NULL);

    frame = P->callback;

    while (frame != NULL) {
        assert(VALID_CALLBACK_FRAME(frame));

        (frame->f)(buf, len, frame->data);
        frame = frame->down;
    }
}


// Miscellaneous other helper functions

// Send the string 'str' to the buf, handling whitespace as required.
//
// If skipping_whitespace is NULL, or *skipping_whitespace is
// SKIPPING_WHITESPACE_NONORM, do not skip spaces.
//
// If *skipping_whitespace is SKIPPING_WHITESPACE_COLLAPSE or _TRUE,
// then do skip leading whitespace, with the value of
// *skipping_whitespace being set as appropriate by this function to
// communicate state to the next invocation.
//
// 'len' gives the number of characters to be output.  If 'len == 0',
// then the string is null-terminated.
//
// Return true if any non-whitespace characters were emitted.
static int output_string_with_whitespace_skipping(Buf* buf,
                                                  const XML_Char* str,
                                                  size_t len,
                                                  char *skipping_whitespace)
{
    int found_non_whitespace = 0; // return value
    const byte* start = (const byte*)str;
    const byte* endp;           // end of the text to be output

    if (len == 0) {
        // text is null-terminated
        for (endp=start; *endp!='\0'; endp++) {
            // nothing
        }
    } else {
        endp = start + len;
    }

    // We are processing an XML_Char, which may not, in principle, be
    // a byte.  However if the expat library has not been compiled
    // with the XML_UNICODE parameter, then the input is encoded in
    // UTF-8.  All we care about, in the input string, is whether the
    // characters are whitespace or not; neither the leading or the
    // trailing UTF-8 bytes can be mistaken for whitespace (they all
    // have the high bit set), so we can therefore treat the 'str'
    // contents as bytes.

    if (skipping_whitespace == NULL
        || *skipping_whitespace == SKIPPING_WHITESPACE_NONORM) {
        // this is in the no-normalization mode
        const byte* p;

        for (p=start; p < endp; p++) {
            switch (*p) {
              case '\n':
              case '\r':
                bufstrcpy(buf, (byte*)(*p == '\n' ? "\\n" : "\\r"));
                break;

              case '\\':
                bufstrcpy(buf, (byte*)"\\\\");
                found_non_whitespace = 1;
                break;

              case 0xc2:
                if (endp - p > 1 && p[1] == 0x85) {
                    // this is U+0085, NEL.  This is not special in
                    // XML 1.0, but is a line-end character in XML
                    // 1.1, so escape it.
                    bufstrcpy(buf, (byte*)"\\u0085");
                    p++;
                } else {
                    found_non_whitespace = 1;
                }
                break;

              case 0xe2:
                if (endp - p > 2 && p[1] == 0x80 && p[2] == 0xa8) {
                    // this is U+2028, which is a line-end character in XML 1.1
                    bufstrcpy(buf, (byte*)"\\u2028");
                    p += 2;
                } else {
                    found_non_whitespace = 1;
                }
                break;

              default:
                found_non_whitespace = 1;
                bufchar(buf, *p);
                break;
            }
        }
    } else {
        // 'words' points to the beginning of an extent between whitespace
        const byte* start = (const byte*)str;
        const byte* words = start;
        const byte* p;
        int last_was_whitespace = 0;

        if (*skipping_whitespace == SKIPPING_WHITESPACE_TRUE) {
            // the immediately preceding invocation of this function left
            // this state value at SKIPPING_WHITESPACE_TRUE, meaning that
            // we're 'currently' skipping whitespace.  So continue to do so.
            while (words < endp) {
                int n = isxmlspace(words);
                if (n) {
                    words += n;
                    last_was_whitespace = 1;
                } else {
                    break;
                }
            }
        }

        p = words;
        while (p < endp) {
            int n = isxmlspace(p);
            if (n) {
                // found whitespace
                last_was_whitespace = 1;
                // copy the 'words' to the buf
                if (p-words > 0) {
                    bufstrncpy(buf, words, p-words);
                }
                // ...jump over this space
                p += n;
                // ...then skip over any remaining space
                while (p < endp) {
                    n = isxmlspace(p);
                    if (n)
                        p += n;
                    else
                        break;
                }
                // ...remember that this is the beginning of a non-whitespace sequence
                words = p;
                // ...and add a single ' ' to the buf
                bufchar(buf, ' ');
            } else {
                found_non_whitespace = 1;
                last_was_whitespace = 0;
                p++;
            }
        }
        if (p-words > 0) {
            // we still have a 'word' left
            bufstrncpy(buf, words, p-words);
        }
        if (last_was_whitespace) {
            // the last character in str was whitespace,
            // which we've skipped, so the next invocation of this
            // function should skip whitespace at the beginning of its
            // buffer, too.
            *skipping_whitespace = SKIPPING_WHITESPACE_TRUE;
        } else {
            // the last character was not whitespace -- any whitespace
            // we encounter when this function is next called should
            // be collapsed rather than skipped.
            *skipping_whitespace = SKIPPING_WHITESPACE_COLLAPSE;
        }
    }

    return found_non_whitespace;
}

/**
 * Returns the version string of the underlying Expat library
 */
const char* xoxa_parser_library_ident(void)
{
    return XML_ExpatVersion();
}
