#ifndef XML_EXPAT_H
#define XML_EXPAT_H 1

#include <stdio.h>
#include <stdint.h>
#include <setjmp.h>

#include <expat.h>

#include "util_types.h"
#include "pgp_types.h"

struct ParserState_s;
/**
 * The structure which contains all of the state of the Xoxa parser.
 */
typedef struct ParserState_s ParserState;



/**
 * The type of the Xoxa callback.
 * The callback takes three arguments:
 *
 *   - 'bytes' is the array of bytes to process.
 *   - 'len' is the number of bytes in this array to process; len may be
 *     passed as zero, indicating that the byte array is null-terminated.
 *   - 'data' is the pointer registered when the callback is registered
 *      with xoxa_push_output_callback.
 */
typedef void (*output_callback_t)(const byte* bytes, size_t len, void* data);

/**
 * The type of the Xoxa callback finaliser.
 * The callback takes two arguments:
 *
 *   - 'bytes' is the array of bytes to process.
 *   - 'data' is the pointer registered when the callback is registered
 *      with xoxa_register_output_callback.
 */
typedef void (*output_callback_end_t)(ParserState*, void* data);

int xoxa_push_output_callback(ParserState* P,
                              output_callback_t cb,
                              void* data,
                              const char* label);

void xoxa_register_output_callback_finalizer(ParserState* P,
                                             output_callback_end_t cb_end);

// public interface

/**
 * Various flags which control the parser.
 * The flags are:
 * <ul>
 *   <li>XOXA_REPORT_ALL  : do not normalize the output
 *   <li>XOXA_QUIET       : be quiet -- not verbose
 *   <li>XOXA_FEEDBACK    : 'normal' feedback
 *   <li>XOXA_CHATTER     : output more extensive tracing information
 * </ul>
 */
typedef uint8_t ParserFlags;
// the flag values are only for communication with the xoxa_parser_flags function
#define XOXA_REPORT_ALL		1<<0
// the following three occupy two bits; XOXA_CHATTER also acts as a mask for these
#define XOXA_QUIET 		1<<1
#define XOXA_FEEDBACK		2<<1
#define XOXA_CHATTER		3<<1
#if XOXA_PRIVATE_INTERFACE
#define SHOWING_MESSAGES(PS_P) ((PS_P)->flags.verbosity > 0)
#define SHOWING_CHATTER(PS_P)  ((PS_P)->flags.verbosity > 1)
#endif

#define INPUT_SOURCE_STRING 0
#define INPUT_SOURCE_FILE 1
typedef struct {
    char type;                  // INPUT_SOURCE_*
    const char* basename;       // the filename, in the case of in.f
    union {
        const char* buffer;
        FILE* f;
    } in;
} InputSource;

struct callback_frame_s {
    output_callback_t f;
    output_callback_end_t f_end;
    unsigned int startdepth;
    void* data;
    const char* label;          // a printable label for this frame
    struct callback_frame_s* down;
};
typedef struct callback_frame_s callback_frame_t;
#define VALID_CALLBACK_FRAME(FRAME) ((FRAME)->f != NULL)


/** The Xoxa parser state structure.
 * Only some of the fields in this structure are documented; these are
 * the only public part of this structure.
 */
struct ParserState_s {
    /** The underlying Expat parser. */
    XML_Parser parser;
    InputSource *src;
    callback_frame_t* callback;
    // element_depth is incremented at the start of any element,
    // and decremented at its end
    unsigned int element_depth;
    Buf* B;
    Buf* text;
    char text_skipping_whitespace; // values of this are private to xml_expat.c
    struct {
        unsigned int report_all    : 1;
        unsigned int verbosity     : 2; // values 0, 1, 2
#define XOXA_DIGEST_TYPE_NONE 0
#define XOXA_DIGEST_TYPE_MD 1
#define XOXA_DIGEST_TYPE_PGP 2
        // also indicates whether pgp or digest is used in union 'd' below
        unsigned int digest_type   : 2;
    } flags;
    union {
        struct {
            void* ctx;
            byte* bytes;
            unsigned int len;
        } digest;
        SignatureDetail pgp;    // if parser_flags & XOXA_DIGEST_IS_PGP_
    } d;
    // following three are filled in from a <?signature?> PI
    const char* signature_armor;
    const char* signature_xpath;
    const char* signature_algorithm;
    output_callback_t signature_callback;
    void* signature_callback_data;
    jmp_buf* error_jmp_buf;
    char* error_message;        // NULL if no error

    // The following extra_* functions are so that we can insert the
    // Xoxa handlers in between the caller and the underlying Expat
    // handlers, and so implement the Expat interface transparently.
    // Each Xoxa handler is written so that it calls the corresponding
    // extra_* handler as well.
    XML_StartElementHandler extra_StartElementHandler;
    XML_EndElementHandler extra_EndElementHandler;
    XML_CharacterDataHandler extra_CharacterDataHandler;
    XML_ProcessingInstructionHandler extra_ProcessingInstructionHandler;
    XML_StartCdataSectionHandler extra_StartCdataSectionHandler;
    XML_EndCdataSectionHandler extra_EndCdataSectionHandler;
    XML_StartNamespaceDeclHandler extra_StartNamespaceDeclHandler;
    XML_EndNamespaceDeclHandler extra_EndNamespaceDeclHandler;
    XML_XmlDeclHandler extra_XmlDeclHandler;
    XML_SkippedEntityHandler extra_SkippedEntityHandler;
    void* extra_UserData;
};

#if XOXA_PRIVATE_INTERFACE
ParserState* xoxa_make_new_parserstate_(InputSource* src,
                                        XML_Parser parser,
                                        ParserState* copy_of);

// assertions for checking the ParserState structure.
// P->src may be NULL, if we're planning to use the GetBuffer
// mechanism to read the content, rather than xoxa_parse_*.
#define PARSERSTATE_OK_TEST_(P)                 \
    ((P) != NULL                                \
     && (P)->parser != NULL                     \
     && (P)->B != NULL                          \
     && (P)->text != NULL)
#define OK_PARSERSTATE(P) do {                  \
        assert(PARSERSTATE_OK_TEST_(P));        \
    } while (0)

// Allocate and fail on any memory errors
#define MALLOC(var, Type) if ((var = (Type*)malloc(sizeof(Type))) == NULL) { \
        fatal("Can't allocate space for " # Type);                      \
    }

#endif /* XOXA_PRIVATE_INTERFACE */

ParserState* xoxa_new_parser(InputSource*);
void xoxa_parser_flags(ParserState*, ParserFlags);
void xoxa_release_parser(ParserState*);

int xoxa_parse(ParserState*);
void xoxa_signature_fd(ParserState*, int);
byte* xoxa_get_digest(ParserState*, size_t*);
SignatureDetail* xoxa_get_signature(ParserState* P);

void xoxa_null_callback(const byte* buf, size_t len, void* data);
void xoxa_default_callback(const byte* buf, size_t len, void* data);

const char* xoxa_error_message(ParserState*);

InputSource* xoxa_new_input_source_from_string(const char* xmlcontent);
InputSource* xoxa_new_input_source_from_filename(const char* filename);
InputSource* xoxa_new_input_source_from_fd(const int fd);

const char* xoxa_parser_library_ident(void);

#endif /* XML_EXPAT_H */
